/********************************************************************
 * Timed Bomb example in C
 * Copyright 2002-2003 by Miro Samek, http://www.quantum-leaps.com 
 * All Rights Reserved.
 *******************************************************************/
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "qassert.h"
#include "fsm.h"
#include "bomb.h"                                /* Bomb resources */

DEFINE_THIS_FILE;

/* Bomb class -----------------------------------------------------*/
typedef struct Bomb Bomb;
struct Bomb {
   Fsm super_;
   int timeout__;
   int code__;
};

Bomb *BombCtor(Bomb *me);
void Bomb_initial(Bomb *me, Event const *e);
void Bomb_setting(Bomb *me, Event const *e);
void Bomb_timing(Bomb *me, Event const *e);
void Bomb_blast(Bomb *me, Event const *e);

/*-----------------------------------------------------------------*/
enum BombSignals {
   ARM_SIG = USER_SIG,
   TICK_SIG,
   UP_SIG,
   DOWN_SIG
}; 

enum {
   INIT_TIMEOUT = 5,
   MAX_TIMEOUT = 10,
   DEFUSE_CODE = 0x0D                               /* 1101 binary */
};

/*-----------------------------------------------------------------*/
static Bomb locApp;
static HINSTANCE locInst;                      /* this locInstance */
static HWND locHwnd;                            /* the main window */
static char const locAppName[] = "Bomb";

/*.................................................................*/
Bomb *BombCtor(Bomb *me) {
   FsmCtor_(&me->super_, &Bomb_initial);
   return me;
}

void Bomb_initial(Bomb *me, Event const *e) {
   me->timeout__ = INIT_TIMEOUT;
   SetDlgItemInt(locHwnd, IDC_TIMEOUT, me->timeout__, FALSE);
   FsmInitialTran_((Fsm *)me, (State)&Bomb_setting);
}

void Bomb_setting(Bomb *me, Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "setting");
      ShowWindow(GetDlgItem(locHwnd, IDC_GREY), TRUE);
      break;
   case UP_SIG:
      if (me->timeout__ < MAX_TIMEOUT) {
         ++me->timeout__;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, me->timeout__, FALSE);
      }
      break;
   case DOWN_SIG:
      if (me->timeout__ > 1) {
         --me->timeout__;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, me->timeout__, FALSE);
      }
      break;
   case ARM_SIG:
      me->code__ = 0;                       /* invalidate the code */
      FsmTran_((Fsm *)me, (State)&Bomb_timing);
      break;
   case EXIT_SIG:
      ShowWindow(GetDlgItem(locHwnd, IDC_GREY), FALSE);
      break;
   }
}

void Bomb_timing(Bomb *me, Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "timing");
      SetTimer(locHwnd, 1, 1000, 0);
      ShowWindow(GetDlgItem(locHwnd, IDC_RED), 
                 (me->timeout__ & 1) == 0);
      ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), 
                 (me->timeout__ & 1) != 0);
      break;
   case TICK_SIG:
      if (me->timeout__ > 0) {
         --me->timeout__;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, me->timeout__, FALSE);
         ShowWindow(GetDlgItem(locHwnd, IDC_RED), 
                    (me->timeout__ & 1) == 0);
         ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), 
                    (me->timeout__ & 1) != 0);
      }
      else {                                    /* timeout expired */
         FsmTran_((Fsm *)me, (State)&Bomb_blast);
      }
      break;
   case UP_SIG:
      me->code__ = (me->code__ << 1) | 1;
      break;
   case DOWN_SIG:
      me->code__ <<= 1;
      break;
   case ARM_SIG:
      if (me->code__ == DEFUSE_CODE) {
         FsmTran_((Fsm *)me, (State)&Bomb_setting);
      } 
      else {
         me->code__ = 0;                         /* clear the code */
      }
      break;
   case EXIT_SIG:
      KillTimer(locHwnd, 1);              /* don't leak the timer! */
      ShowWindow(GetDlgItem(locHwnd, IDC_RED), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), FALSE);
      break;
   }
}

void Bomb_blast(Bomb *me, Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "blast");
      Beep(1000, 20);                           /* make some noise */
      ShowWindow(GetDlgItem(locHwnd, IDC_ARM), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_UP), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_DOWN), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_BLAST), TRUE);
      break;
   }
}

/*.................................................................*/
void onAssert__(char const *file, unsigned line) {
   char str[160];
   sprintf(str, "Assertion failed in %s, line %d", file, line);        
   MessageBox(locHwnd, str, locAppName, MB_ICONEXCLAMATION | MB_OK);
   exit(-1);
}
/*-----------------------------------------------------------------*/
BOOL CALLBACK BombDlg(HWND hWnd, UINT iMsg, 
                      WPARAM wParam, LPARAM lParam)
{
   BOOL isHandled = TRUE;
   Event e;
   e.sig = iMsg;
   switch (iMsg) {
   case WM_INITDIALOG:
      locHwnd = hWnd;
      FsmInit((Fsm *)&locApp, 0);      /* take the initial transition */
      return TRUE;
   case WM_TIMER:
      e.sig = TICK_SIG;
      /* intentionally fall thru */
   case WM_COMMAND:                     
      switch (LOWORD(wParam)) {
      case IDCANCEL:
         EndDialog(locHwnd, 0);
         break;
      case IDC_ARM:
         e.sig = ARM_SIG;
         break;
      case IDC_UP:
         e.sig = UP_SIG;
         break;
      case IDC_DOWN:
         e.sig = DOWN_SIG;
         break;
      default:
         isHandled = FALSE;
         break;
      }      
      FsmDispatch((Fsm *)&locApp, &e);          /* dispatch the event */
      return isHandled;
   }
   return FALSE;
}
/*.................................................................*/
int WINAPI WinMain(HINSTANCE hlocInst, HINSTANCE hPrevlocInst,
                   PSTR cmdLine, int iCmdShow)
{
   InitCommonControls();           /* load common controls library */
   locInst = hlocInst;                 /* store locInstance handle */
   BombCtor(&locApp);         /* construct the Bomb locApplication */
   DialogBox(hlocInst, MAKEINTRESOURCE(IDD_DIALOG), NULL, BombDlg);                  
   return 0;        /* exit locApplication when the dialog returns */
}

