/////////////////////////////////////////////////////////////////////
// Timed Bomb example in C++
// Copyright 2002-2003 by Miro Samek, http://www.quantum-leaps.com 
// All Rights Reserved.
/////////////////////////////////////////////////////////////////////
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "qassert.h"
#include "fsm.h"
#include "bomb.h"                                   // Bomb resources

DEFINE_THIS_FILE;

// Bomb class -------------------------------------------------------
class Bomb : public Fsm {
   int myTimeout;
   int myCode;
public:
   Bomb() : Fsm((State)&Bomb::initial) {}
   void initial(Event const *e);
   void setting(Event const *e);
   void timing(Event const *e);
   void blast(Event const *e);

   friend BOOL CALLBACK BombDlg(HWND hwnd, UINT iEvt, 
                                WPARAM wParam, LPARAM lParam);
};

//-------------------------------------------------------------------
enum BombSignals {
   ARM_SIG = USER_SIG,
   TICK_SIG,
   UP_SIG,
   DOWN_SIG
}; 

enum {
   INIT_TIMEOUT = 5,
   MAX_TIMEOUT = 10,
   DEFUSE_CODE = 0x0D                                  // 1101 binary
};

//-------------------------------------------------------------------
static Bomb locApp;
static HINSTANCE locInst;                         // this locInstance
static HWND locHwnd;                               // the main window
static char locAppName[] = "Bomb";

//...................................................................
void Bomb::initial(Event const *e) {
   myTimeout = INIT_TIMEOUT;
   SetDlgItemInt(locHwnd, IDC_TIMEOUT, myTimeout, FALSE);
   initialTran((State)&Bomb::setting);
}
//...................................................................
void Bomb::setting(Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "setting");
      ShowWindow(GetDlgItem(locHwnd, IDC_GREY), TRUE);
      break;
   case UP_SIG:
      if (myTimeout < MAX_TIMEOUT) {
         ++myTimeout;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, myTimeout, FALSE);
      }
      break;
   case DOWN_SIG:
      if (myTimeout > 1) {
         --myTimeout;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, myTimeout, FALSE);
      }
      break;
   case ARM_SIG:
      myCode = 0;                              // invalidate the code
      tran((State)&Bomb::timing);
      break;
   case EXIT_SIG:
      ShowWindow(GetDlgItem(locHwnd, IDC_GREY), FALSE);
      break;
   }
}
//...................................................................
void Bomb::timing(Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "timing");
      SetTimer(locHwnd, 1, 1000, 0);
      ShowWindow(GetDlgItem(locHwnd, IDC_RED), 
                 (myTimeout & 1) == 0);
      ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), 
                 (myTimeout & 1) != 0);
      break;
   case TICK_SIG:
      if (myTimeout > 0) {
         --myTimeout;
         SetDlgItemInt(locHwnd, IDC_TIMEOUT, myTimeout, FALSE);
         ShowWindow(GetDlgItem(locHwnd, IDC_RED), 
                    (myTimeout & 1) == 0);
         ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), 
                    (myTimeout & 1) != 0);
      }
      else {                                       // timeout expired
         tran((State)&Bomb::blast);
      }
      break;
   case UP_SIG:
      myCode = (myCode << 1) | 1;
      break;
   case DOWN_SIG:
      myCode <<= 1;
      break;
   case ARM_SIG:
      if (myCode == DEFUSE_CODE) {
         tran((State)&Bomb::setting);
      } 
      else {
         myCode = 0;                                // clear the code
      }
      break;
   case EXIT_SIG:
      KillTimer(locHwnd, 1);                 // don't leak the timer!
      ShowWindow(GetDlgItem(locHwnd, IDC_RED), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_YELLOW), FALSE);
      break;
   }
}
//...................................................................
void Bomb::blast(Event const *e) {
   switch (e->sig) {
   case ENTRY_SIG:
      SetDlgItemText(locHwnd, IDC_STATE, "blast");
      Beep(1000, 20);                              // make some noise
      ShowWindow(GetDlgItem(locHwnd, IDC_ARM), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_UP), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_DOWN), FALSE);
      ShowWindow(GetDlgItem(locHwnd, IDC_BLAST), TRUE);
      break;
   }
}
//...................................................................
void onAssert__(char const *file, unsigned line) {
   char str[160];
   sprintf(str, "Assertion failed in %s, line %d", file, line);        
   MessageBox(locHwnd, str, locAppName, MB_ICONEXCLAMATION | MB_OK);
   exit(-1);
}
//-------------------------------------------------------------------
BOOL CALLBACK BombDlg(HWND hwnd, UINT iMsg, 
                                   WPARAM wParam, LPARAM lParam)
{
   BOOL isHandled = TRUE;
   Event e;
   e.sig = iMsg;
   switch (iMsg) {
   case WM_INITDIALOG:
      locHwnd = hwnd;
      locApp.init(0);                  // take the initial transition
      return TRUE;
   case WM_TIMER:
      e.sig = TICK_SIG;
      // intentionally fall thru
   case WM_COMMAND:                     
      switch (LOWORD(wParam)) {
      case IDCANCEL:
         EndDialog(hwnd, 0);
         break;
      case IDC_ARM:
         e.sig = ARM_SIG;
         break;
      case IDC_UP:
         e.sig = UP_SIG;
         break;
      case IDC_DOWN:
         e.sig = DOWN_SIG;
         break;
      default:
         isHandled = FALSE;
         break;
      }      
      locApp.dispatch(&e);                      // dispatch the event
      return isHandled;
   }
   return FALSE;
}
//...................................................................
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst,
                   PSTR cmdLine, int iCmdShow)
{
   InitCommonControls();              // load common controls library
   locInst = hInst;                       // store locInstance handle
   DialogBox(hInst, MAKEINTRESOURCE(IDD_DIALOG), NULL, BombDlg);                  
   return 0;              // exit application when the dialog returns
}

