// This is the sample code for the "Hough Transform" article
// in the November 2003 article of the C/C++ User Journal
//
// written by Andrew Queisser (andrew.queisser@hp.com)

#include <vector>
#include <math.h>
#include <iostream>
#include "houghtrans.h"

using namespace std;


HoughAccum::HoughAccum() : m_dx(0), m_dy(0)
{
}

HoughAccum::~HoughAccum()
{
}

void HoughAccum::Clear()
{
  m_accum.assign(m_accum.size(), 0);
  m_numAccumulated = 0;
}

// This is the template method AddPoint. Function is public
// and non-virtual in the base class so the limit checking
// can't be circumvented. Actual work gets done in DoAddPoint
bool HoughAccum::AddPoint(int x, int y)
{
  if (x>m_dx || y>m_dy || x<0.0 || y<0.0)
    return false;
  return DoAddPoint(x, y);
}

// Constructor sizes the sine and cosine tables according to
// the numAngleCells parameter.
LineHoughAccum::LineHoughAccum(int numAngleCells) : HoughAccum()
{
  m_cosTable.resize(numAngleCells);
  m_sinTable.resize(numAngleCells);
}

LineHoughAccum::~LineHoughAccum()
{
}

/* Init sizes the accumulator. The range for the column dimension
(rho) depends on dx and dy so the accumulator can be sized to
cover all possible rho values and all possible angle values.
*/
void LineHoughAccum::Init(int dx, int dy)
{
  if (dx == m_dx && dy == m_dy)
    return;
  m_dx = dx;
  m_dy = dy;
  int numAngleCells = m_cosTable.size();
  m_rowLength = 2 * (int)sqrt(dx*dx + dy*dy);
  m_accum.resize(numAngleCells*m_rowLength);
  // Calculate the sine/cosine LUTs with a scaling of 1024 for
  // easy downscaling with a right-shift of 10.
  int angleIndex;
  double a, angleStep = PI/numAngleCells;
  for (a=0.0, angleIndex=0; angleIndex<m_cosTable.size(); angleIndex++)
  {
    m_cosTable[angleIndex] = 1024*cos(a);
    m_sinTable[angleIndex] = 1024*sin(a);
    a += angleStep;
  }
  Clear();
}

/* This is the main function to calculate the Hough Transform
for line detection. Loop through all angle values a, calculate the
distance value rho(a) and accumulate.

The accumulation is complicated by the fact that we can't just accumulate
m_accum[a*m_rowLength + cos(rho)*x + sin(rho)*y] but have to fill in 
areas where more than one distance value is filled in for an angle value.
*/
bool LineHoughAccum::DoAddPoint(int x, int y)
{
  int a;
  int distCenter = m_rowLength/2;

  // prime the lastDist variable so our linear fill works below
  int lastDist = distCenter + ((m_cosTable[0]*x + m_sinTable[0]*y) >> 10);

  // loop through all angles a from 0 to 180 degrees
  for (a=1; a<m_cosTable.size(); a++)
  {
    // Calculate the basic line equation dist = cos(a)*x + sin(a)*y.
    // Shift to center of row to cover negative values
    int dist = distCenter + ((m_cosTable[a]*x + m_sinTable[a]*y) >> 10);
    // sanity check to make sure we stay within our row
    if (dist >= 0 && dist<m_rowLength)
    {
      // fill in all values in row a, not just a single cell
      int stepDir = dist>lastDist ? 1 : -1;
      for (int cell=lastDist; cell!=dist; cell+=stepDir)
        m_accum[a*m_rowLength + cell] += 1;
    }
    lastDist = dist;
  }
  m_numAccumulated += 1;

  return true;
}

// main driver test program, uses external functions LoadBMPFile
// and SaveBMPFile to fill in an array of 8bit pixels. Note that the
// BMP functions can only deal with monochrome 8bit bitmaps where
// any pixel other than 0 is considered an edge pixel that will be
// accumulated.
int main(int argc, char* argv[])
{
  if (argc < 2)
  {
    cout << "usage: " << argv[0] << " inputbitmapfile\n\nwhere inputbitmapfile is a 8 bit binarized bitmap\n";
    exit(-1);
  }

  int dx,dy;
  int x, y;
  unsigned char *points;  // LoadBMPFile will allocate pixels here
  LineHoughAccum ha(360); // reservere 360 angle values (0.5 degree resolution)

  extern unsigned char *LoadBMPFile(const char *f, int *dx, int *dy);
  extern void SaveBMPFile(const char *f, unsigned char *pix, int dx, int dy);

  // use your favorite image reading routine to fill in the points array
  points = LoadBMPFile(argv[1], &dx, &dy);

  cout << "starting transform\n";

  if (points)
  {
    ha.Init(dx, dy);
    for (y=0; y<dy; y++)
      for (x=0; x<dx; x++)
        if (points[y*dx + x])
          ha.AddPoint(x,y);
  }

  cout << "Transform done. Number of points: " << ha.NumAccumulated() << endl;

  ////////////////////////////////////////////////////////////////
  // build an image from the Hough accumulator and the
  // resulting lines
  int accDx, accDy;
  ha.GetAccumSize(&accDy, &accDx);
  int outputSize = accDx*accDy;
  unsigned char *outPix = new unsigned char [accDx*accDy];

  // Build an image of the accumulator where the maximum
  // value is scaled to 255, first we need to find the max
  int cell, maxCell=0;
  for (y=0; y<accDy; y++)
    for (x=0; x<accDx; x++)
    {
      cell = ha.GetCell(y,x);
      if (cell > maxCell)
        maxCell = cell;
    }

  // loop through accumulator and scale each value
  for (y=0; y<accDy; y++)
    for (x=0; x<accDx; x++)
    {
      int pix = (255*ha.GetCell(y,x))/maxCell;
      outPix[y*accDx + x] = pix;
    }
  // write the accumulator to a .bmp file for viewing
  SaveBMPFile("houghaccum.bmp", outPix, accDx, accDy);
  delete [] outPix;
  //////////////////////////////////////////////////////////////////

  //////////////////////////////////////////////////////////////////
  // Now we want to generate an image that contains some lines 
  // overlayed over the original image
  unsigned char *linePic = new unsigned char [dx*dy];
  memset(linePic,0,dx*dy);
  // first copy the original image into the linePic with each
  // original image set to 128. Lines will be overlayed with 255
  int i;
  for (i=0; i<dx*dy; i++)
    linePic[i] = points[i]!=0 ? 128 : 0;

  int thresh = 50;  // this is the minimum number of hits in the
                    // accumulator we will consider. This value
                    // is roughly equivalent to the length of the line
  const int numLinesToFind = 15;
  for (i = 0; i < numLinesToFind; i++)
  {
    // find the next maximum
    int maxCell = 0;
    int xMax, yMax;
    for (y=0; y<accDy; y++)
    {
      for (x=0; x<accDx; x++)
      {
        if (ha.GetCell(y, x) > maxCell) 
        {
          maxCell = ha.GetCell(y, x);
          xMax = x;
          yMax = y;
        }
      }
    }
    // we're done when biggest maximum is smaller than thresh
    if (maxCell < thresh)
      break;

    // get the line equation and print it out
    int rho;
    double theta;
    ha.GetEquation(yMax, xMax, &rho, &theta);
    cout << "Line at angle " << yMax << " degrees, distance " << rho << endl;

    // Draw line, brute force, just check whether a pixel lies on the
    // line or not. Could be done much, much more efficiently
    for (y=0; y<dy; y++)
      for (x=0; x<dx; x++)
      {
        if (fabs(x*cos(theta)+y*sin(theta)-rho) < 1.0)
          linePic[y*dx+x] = 255;
      }

    // Now we need to clear out the neighborhood of the peak we just
    // found. In this case we erase a range of 10 pixel distance and
    // 5 degrees angular range
    int xClear = 10;
    int yClear = 10;
    // zero out neighborhood of line, make sure we don't fall of the
    // edge of the array
    int yClearStart = yMax-yClear;
    if (yClearStart < 0)
      yClearStart = 0;
    int yClearEnd = yMax+yClear;
    if (yClearEnd >= accDy)
      yClearEnd = accDy-1;

    int xClearStart = xMax-xClear;
    if (xClearStart < 0)
      xClearStart = 0;
    int xClearEnd = xMax+xClear;
    if (xClearEnd >= accDx)
      xClearEnd = accDx-1;

    for (y=yClearStart; y<=yClearEnd; y++)
      for (x=xClearStart; x<=xClearEnd; x++)
        ha.SetCell(y,x,0);
  }

  SaveBMPFile("lines.bmp",linePic,dx,dy);

  delete [] points;
  delete [] linePic;

  return 0;
}

