/* Base class for Hough Transform accumulator. Defines storage
and basic interface to add points and inspect the accumulator
results. DoAddPoint is a template method that will be defined
by derived classes.

Note: additional speed gain can be achieved by making m_accum
an actual two-dim array with reasonable upper limits.
*/
#include <vector>
const double PI=3.141592653;

class HoughAccum
{
protected:
  int m_dx, m_dy;
  std::vector<int> m_accum;  // col=rho, row=theta
  int m_rowLength;
  int m_numAccumulated;
  virtual bool DoAddPoint(int x, int y)=0;
public:
  HoughAccum();
  virtual ~HoughAccum();
  // Abstract method Init does everything necessary to start
  // processing images of size dx*dy
  virtual void Init(int dx, int dy)=0;
  // Clear empties the accumulator. Call before processing a
  // each new image
  virtual void Clear();
  // AddPoint is called once for each edge point added to the
  // accumulator
  bool AddPoint(int x, int y);
  // GetCell/SetCell are accessors to get results out of the
  // accumulator
  int GetCell(int row, int col) { return m_accum[row*m_rowLength + col]; }
  void SetCell(int row, int col, int value) { m_accum[row*m_rowLength + col] = value; }
  // GetAccumSize returns the dimensions of the accumulator.
  // Needed to calculate the line equations from peaks in the
  // accumulator
  void GetAccumSize(int *numRows, int *numCols) 
  { 
    *numRows = m_accum.size()/m_rowLength;
    *numCols  = m_rowLength;
  }
  // NumAccumulated simply returns the number of times AddPoint
  // was called
  int NumAccumulated() { return m_numAccumulated; }
};


// LineHoughAccum is the implementation for a linear Hough
// Transform accumulator.
class LineHoughAccum : public HoughAccum
{
protected:
  // Integer lookup tables for sines and cosines times
  // 1024 for quick scaling
  std::vector<int> m_cosTable;
  std::vector<int> m_sinTable;
  // AddPoint is te actual line Hough Transform function
  virtual bool DoAddPoint(int x, int y);
public:
  // LineHoughAccum needs to be told how many cells to reserve
  // for angles between 0 and 180 degrees
  LineHoughAccum(int numAngleCells);
  virtual ~LineHoughAccum();
  // Init sets up the sin and cosine lookup tables and sizes
  // the accumulator. Init needs to be called once at startup
  // and again when the image size changes
  virtual void Init(int dx, int dy);
  // Function to get line equation from accumulator cell
  virtual void GetEquation(int row, int col, 
                           int *rho, double *theta)
  {
    *theta = (PI*row)/m_cosTable.size();
    *rho   = col - m_rowLength/2;
  }
};

