/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_auto_buffer.h (originally MTLocBfr.h, ::SynesisStl)
 *
 * Purpose:     Contains the auto_buffer template class.
 *
 * Created:     19th January 2002
 * Updated:     9th September 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER
#define _STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#define _STLSOFT_VER_H_STLSOFT_AUTO_BUFFER_MAJOR     1
#define _STLSOFT_VER_H_STLSOFT_AUTO_BUFFER_MINOR     11
#define _STLSOFT_VER_H_STLSOFT_AUTO_BUFFER_REVISION  5
#define _STLSOFT_VER_H_STLSOFT_AUTO_BUFFER_EDIT      53
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
 #include "stlsoft.h"                   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */
#ifdef _STLSOFT_AUTO_BUFFER_ALLOW_UDT
# define _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD
# pragma message("_STLSOFT_AUTO_BUFFER_ALLOW_UDT is deprecated. Use _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD instead")
#endif /* _STLSOFT_AUTO_BUFFER_ALLOW_UDT */
#ifndef _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD
 #ifndef _STLSOFT_INCL_H_CONSTRAINTS
  #include "stlsoft_constraints.h"      // Include the STLSoft constraints header
 #endif /* !_STLSOFT_INCL_H_CONSTRAINTS */
#endif /* _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class auto_buffer
//
/// \brief This class provides an efficient variable automatic buffer
/// 
/// This class provides an efficient replacement for dynamic memory block
/// allocation when the block size generally falls under a certain limit. In
/// such cases, significant performance benefits can be achieved by using an
/// instance of a parameterisation of auto_buffer, whose size parameter SPACE
/// is set to a level to cater for most of the requested sizes. Only where the
/// size of the buffer needs to be larger than this limit does an allocation
/// occur from the heap/free-store via the given allocator.
/// 
/// \note The buffer works correctly whether the given allocator throws an exception on allocation failure, or returns NULL.
/// 
/// \param T The type of the elements in the array
/// \param A The allocator type
/// \param SPACE The number of elements in the array. With translators that support default template arguments, this is defaulted to 512

template<   ss_typename_param_k T
        ,   ss_typename_param_k A
#ifdef __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT
        ,   ss_size_t           SPACE = 512
#else
        ,   ss_size_t           SPACE
#endif /* __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT */
        >
class auto_buffer
#ifndef __STLSOFT_COMPILER_IS_BORLAND
    : protected A
#endif /* !__STLSOFT_COMPILER_IS_BORLAND */
{
public:
    /// The value type
    typedef T                       value_type;
    /// The allocator type
    typedef A                       allocator_type;
    enum
    {
        /// The number of items in the internal buffer
        space = SPACE
    };
    /// The size type
    typedef ss_size_t                 size_type;
    /// The type of the current parameterisation
    typedef auto_buffer<T, A, SPACE>  class_type;
    /// The pointer type
    typedef value_type                *pointer;
    /// The non-mutable (const) pointer type
    typedef value_type const          *const_pointer;
    /// The iterator type
    typedef value_type                *iterator;
    /// The non-mutable (const) iterator type
    typedef value_type const          *const_iterator;

// Implementation
private:
#ifdef __STLSOFT_COMPILER_IS_BORLAND
    static allocator_type &_get_ator()
    {
        static allocator_type   s_allocator;

        return s_allocator;
    }
#else
    allocator_type &_get_ator()
    {
        return *this;
    }
#endif /* __STLSOFT_COMPILER_IS_BORLAND */

// Construction
public:
    /// \brief Constructs an auto_buffer with the given number of elements
    ///
    /// Constructs an auto_buffer with the given number of elements. If the 
    /// allocation fails by throwing an exception, that exception is passed
    /// through to the caller. If allocation fails by returning a null 
    /// pointer the auto_buffer instance is correctly constructed, and the
    /// \link #size size() \endlink method returns 0.
    ///
    /// \see \link #size size() \endlink
    /// \param cItems The number of items in the constructed instance
    ss_explicit_k auto_buffer(size_type cItems)
#ifdef __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT
        : m_buffer((space < cItems) ? _get_ator().allocate(cItems, 0) : m_internal)
#else
        : m_buffer((space < cItems) ? _get_ator().allocate(cItems) : m_internal)
#endif /* __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT */
        , m_cItems((m_buffer != 0) ? cItems : 0)
    {
        // Can't create one with an empty buffer. Though such is not legal
        // it is supported by some compilers, so we must ensure it cannot be
        // so
        stlsoft_static_assert(space != 0);

        // These assertions ensure that the member ordering is not
        // changed, invalidating the initialisation logic of m_buffer and
        // m_cItems. The runtime assert is included for those compilers that
        // do not implement compile-time asserts.
        stlsoft_static_assert(stlsoft_raw_offsetof(class_type, m_buffer) < stlsoft_raw_offsetof(class_type, m_cItems));
#if defined(__STLSOFT_COMPILER_IS_WATCOM)
        stlsoft_assert(stlsoft_raw_offsetof(class_type, m_buffer) < stlsoft_raw_offsetof(class_type, m_cItems));
#else
        stlsoft_message_assert("m_buffer must be before m_cItems in the auto_buffer definition", stlsoft_raw_offsetof(class_type, m_buffer) < stlsoft_raw_offsetof(class_type, m_cItems));
#endif /* !__STLSOFT_COMPILER_IS_WATCOM */

#ifndef _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD
        // Use the type_is_non_class_or_trivial_class constraint to ensure that
        // no type is managed in auto_buffer which would result in
        // dangerous mismanagement of the lifetime of its instances.
        //
        // Preprocessor specification of _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD
        // prevents this, but the natural rules of the language will
        // still prevent non POD types being placed in m_internal[].
        stlsoft_constraint_type_is_non_class_or_trivial_class(value_type);
#endif /* _STLSOFT_AUTO_BUFFER_ALLOW_NON_POD */
    }
    /// \brief Releases the allocated element array
    ///
    /// Releases any allocated memory. If the internal memory buffer was 
    /// used, then nothing is done, otherwise the allocated memory is
    /// returned to the allocator.
#ifdef __STLSOFT_COMPILER_IS_BORLAND
    ~auto_buffer()
#else
    ~auto_buffer() stlsoft_throw_0()
#endif /* __STLSOFT_COMPILER_IS_BORLAND */
    {
        if(space < m_cItems)
        {
            stlsoft_assert(m_buffer != 0);

#ifdef __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
            _get_ator().deallocate(m_buffer, m_cItems);
#else
            _get_ator().deallocate(m_buffer);
#endif /* __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT */
        }
    }

// Operators
public:
    // Note: The following two const and non-const implicit conversion
    // operators are correctly implemented. However, GCC will pedantically
    // give a verbose warning describing its having selected one over the
    // other, and this is, in current versions of the compiler, not
    // suppressable. The warnings must, alas, simply be ignored.

    /// \brief An implicit conversion to a pointer to the start of the element array
    operator pointer ()
    {
        return m_buffer;
    }

#ifdef _STLSOFT_AUTO_BUFFER_ALLOW_CONST_CONVERSION_OPERATOR
    /// \brief An implicit conversion to a pointer-to-const to the start of the element array
    operator const_pointer () const
    {
        return m_buffer;
    }
#endif /* _STLSOFT_AUTO_BUFFER_ALLOW_CONST_CONVERSION_OPERATOR */

// Accessors
public:
    /// \brief Returns a pointer to the element array
    pointer data()
    {
        return m_buffer;
    }
    /// \brief Returns a pointer-to-const to the element array
    const_pointer data() const
    {
        return m_buffer;
    }

// Iteration
public:
    /// \brief Returns a non-mutating iterator representing the start of the sequence
    const_iterator begin() const
    {
        return m_buffer;
    }
    /// \brief Returns a non-mutating iterator representing the end of the sequence
    ///
    /// \note In the case where memory allocation has failed in the context
    /// where exceptions are not thrown for allocation failure, this method will
    /// return the same value as begin(). Hence, operations on the <i>empty</i>
    /// auto_buffer<> instance will be safe if made in respect of the range
    /// defined by [begin(), end()).
    const_iterator end() const
    {
        return m_buffer + m_cItems;
    }

    /// \brief Returns a mutable iterator representing the start of the sequence
    iterator begin()
    {
        return m_buffer;
    }
    /// \brief Returns a mutable iterator representing the end of the sequence
    ///
    /// \note In the case where memory allocation has failed in the context
    /// where exceptions are not thrown for allocation failure, this method will
    /// return the same value as begin(). Hence, operations on the <i>empty</i>
    /// auto_buffer<> instance will be safe if made in respect of the range
    /// defined by [begin(), end()).
    iterator end()
    {
        return m_buffer + m_cItems;
    }

// Attributes
public:
    /// \brief Returns the number of elements in the auto_buffer
    ///
    /// \note In the case where memory allocation has failed in the context
    /// where exceptions are not thrown for allocation failure, this method will
    /// return 0. Hence, operations on the <i>empty</i> auto_buffer<> instance
    /// will be safe if made in respect of the value returned by this method.
    size_type size() const
    {
        return m_cItems;
    }

    /// \brief Indicates whether the buffer has any contents
    ///
    /// \note This will only ever be true when an allocation above the number
    /// of elements in the internal array has been requested, and failed.
    ss_bool_t empty() const
    {
        return 0 == m_cItems;
    }

// Members
private:
    value_type *const   m_buffer;           // Pointer to used buffer
    size_type const     m_cItems;           // Number of items in buffer
    value_type          m_internal[space];  // Internal storage

// Not to be implemented
private:
    auto_buffer(class_type const &rhs);
    auto_buffer const &operator =(class_type const &rhs);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Shims
 */

#ifndef __STLSOFT_CF_TEMPLATE_SHIMS_NOT_SUPPORTED

template<   ss_typename_param_k T
        ,   ss_typename_param_k A
        ,   ss_size_t SPACE
        >
inline ss_bool_t is_empty(auto_buffer<T, A, SPACE> const &b)
{
    return b.empty();
}

#endif /* !__STLSOFT_CF_TEMPLATE_SHIMS_NOT_SUPPORTED */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER */

/* ////////////////////////////////////////////////////////////////////////// */
