/* /////////////////////////////////////////////////////////////////////////////
 * File:        unixstl_performance_counter.h
 *
 * Purpose:     performance_counter class.
 *
 * Created:     16th January 2002
 * Updated:     2nd July 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/unixstl
 *                          http://www.unixstl.org/
 *
 *              email:      submissions@unixstl.org  for submissions
 *                          admin@unixstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _INCL_UNIXSTL_H_UNIXSTL_PERFORMANCE_COUNTER
#define _INCL_UNIXSTL_H_UNIXSTL_PERFORMANCE_COUNTER

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#define _UNIXSTL_VER_H_UNIXSTL_PERFORMANCE_COUNTER_MAJOR        1
#define _UNIXSTL_VER_H_UNIXSTL_PERFORMANCE_COUNTER_MINOR        2
#define _UNIXSTL_VER_H_UNIXSTL_PERFORMANCE_COUNTER_REVISION     1
#define _UNIXSTL_VER_H_UNIXSTL_PERFORMANCE_COUNTER_EDIT         24
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _INCL_UNIXSTL_H_UNIXSTL
 #include "unixstl.h"   // Include the UNIXSTL root header
#endif /* !_INCL_UNIXSTL_H_UNIXSTL */

#include <sys/time.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The UNIXSTL components are contained within the unixstl namespace. This is
 * actually an alias for stlsoft::unixstl_project,
 *
 * The definition matrix is as follows:
 *
 * _STLSOFT_NO_NAMESPACE    _UNIXSTL_NO_NAMESPACE   unixstl definition
 * ---------------------    ---------------------   -----------------
 *  not defined              not defined             = stlsoft::unixstl_project
 *  not defined              defined                 not defined
 *  defined                  not defined             unixstl
 *  defined                  defined                 not defined
 *
 */

/* No STLSoft namespaces means no UNIXSTL namespaces */
#ifdef _STLSOFT_NO_NAMESPACES
 #define _UNIXSTL_NO_NAMESPACES
#endif /* _STLSOFT_NO_NAMESPACES */

/* No UNIXSTL namespaces means no unixstl namespace */
#ifdef _UNIXSTL_NO_NAMESPACES
 #define _UNIXSTL_NO_NAMESPACE
#endif /* _UNIXSTL_NO_NAMESPACES */

#ifndef _UNIXSTL_NO_NAMESPACE
 #ifdef _STLSOFT_NO_NAMESPACE
/* There is no stlsoft namespace, so must define ::unixstl */
namespace unixstl
{
 #else
/* Define stlsoft::unixstl_project */

namespace stlsoft
{

namespace unixstl_project
{

 #endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_UNIXSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_performance Performance Library
/// \ingroup libraries
/// \brief This library provides performance measuring facilities 

/// \weakgroup unixstl_perf_library Performance Library (UNIXSTL)
/// \ingroup UNIXSTL libraries_performance
/// \brief This library provides performance measuring facilities for UNIX
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief A performance counter using \c gettimeofday()
///
/// This class provides performance monitoring functionality based around the
/// UNIX \c gettimeofday() API.

class performance_counter
{
private:
    typedef us_sint64_t         epoch_type;
public:
    typedef performance_counter class_type;
    /// \brief The interval type
    ///
    /// The type of the interval measurement, a 64-bit signed integer
    typedef us_sint64_t         interval_type;

// Construction
public:

// Operations
public:
    /// \brief Starts measurement
    ///
    /// Begins the measurement period
    void start();
    /// \brief Ends measurement
    ///
    /// Ends the measurement period
    void stop();

// Attributes
public:
    /// \brief The elapsed count in the measurement period
    ///
    /// This represents the extent, in arbitrary units, of the measurement period
    interval_type   get_period_count() const;
    /// \brief The number of whole seconds in the measurement period
    ///
    /// This represents the extent, in whole seconds, of the measurement period
    interval_type   get_seconds() const;
    /// \brief The number of whole milliseconds in the measurement period
    ///
    /// This represents the extent, in whole milliseconds, of the measurement period
    interval_type   get_milliseconds() const;
    /// \brief The number of whole microseconds in the measurement period
    ///
    /// This represents the extent, in whole microseconds, of the measurement period
    interval_type   get_microseconds() const;

// Members
private:
    typedef struct timeval timeval_t;

    timeval_t   m_start;
    timeval_t   m_end;
};

/* //////////////////////////////////////////////////////////////////////// */

inline void performance_counter::start()
{
    struct timezone tz;

    gettimeofday(&m_start, &tz);

    m_end   =   m_start;
}

inline void performance_counter::stop()
{
    struct timezone tz;

    gettimeofday(&m_end, &tz);
}

inline performance_counter::interval_type performance_counter::get_period_count() const
{
    return get_microseconds();
}

inline performance_counter::interval_type performance_counter::get_seconds() const
{
    interval_type   start   =   m_start.tv_sec + m_start.tv_usec / (1000 * 1000);
    interval_type   end     =   m_end.tv_sec   + m_end.tv_usec   / (1000 * 1000);

    return end - start;
}

inline performance_counter::interval_type performance_counter::get_milliseconds() const
{
    interval_type   start   =   m_start.tv_sec * 1000 + m_start.tv_usec / 1000;
    interval_type   end     =   m_end.tv_sec   * 1000 + m_end.tv_usec   / 1000;

    return end - start;
}

inline performance_counter::interval_type performance_counter::get_microseconds() const
{
    interval_type   start   =   m_start.tv_sec * 1000 * 1000 + m_start.tv_usec;
    interval_type   end     =   m_end.tv_sec   * 1000 * 1000 + m_end.tv_usec;

    return end - start;
}

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group unixstl_perf_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _UNIXSTL_NO_NAMESPACE
 #ifdef _STLSOFT_NO_NAMESPACE
} // namespace unixstl
 #else
} // namespace unixstl_project
} // namespace stlsoft
 #endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_UNIXSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_INCL_UNIXSTL_H_UNIXSTL_PERFORMANCE_COUNTER */

/* ////////////////////////////////////////////////////////////////////////// */
