//-----------------------------------------------------------------------------
// File:  ipaddr.hh
// Date:  Tue Dec 10 22:54:28 PST 2002 
// Author: Martin Casado, Norman Franke
//
// Description:
//
// Yet another ip address class
//
// ----------------------------------------------------------------------------
//
//    Copyright 2003. The Regents of the University of California.
//    All rights reserved.
//
//    This work was produced at the University of California, Lawrence Livermore
//    National Laboratory (UC LLNL) under contract no. W-7405-ENG-48 (Contract
//    48) between the U.S. Department of Energy (DOE) and The Regents of the
//    University of California (University) for the operation of UC LLNL.
//    Copyright is reserved to the University for purposes of controlled
//    dissemination, commercialization through formal licensing, or other
//    disposition under terms of Contract 48; DOE policies, regulations and
//    orders; and U.S. statutes. The rights of the Federal Government are
//    reserved under Contract 48 subject to the restrictions agreed upon by the
//    DOE and University as allowed under DOE Acquisition Letter 91-7.
//
//    DISCLAIMER
//    This software was prepared as an account of work sponsored by an
//    agency of the United States Government. Neither the United States
//    Government nor the University of California nor any of their employees,
//    makes any warranty, express or implied, or assumes any liability or
//    responsibility for the accuracy, completeness, or usefulness of any
//    information, apparatus, product, or process disclosed, or represents that
//    its use would not infringe protectedly-owned rights. Reference herein to any
//    specific commercial products, process, or service by trade name, trademark,
//    manufacturer, or otherwise, does not necessarily constitute or imply its
//    endorsement, recommendation, or favoring by the United States Government or
//    the University of California. The views and opinions of authors expressed
//    herein do not necessarilINFO_SOURCE_ENTITY_IDose of the United States
//    Government or the University of California, and shall not be used for
//    advertising or product endorsement purposes.
//
//
//-----------------------------------------------------------------------------

#ifndef ipaddr_HH
#define ipaddr_HH

#include <string>
#include <cstdio>
#include <cassert>
#include <cstring>
#include <iostream>
#include <stdexcept>

extern "C"
{
#include <netdb.h>
#include <inttypes.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
}

namespace pacman
{

struct ipaddr
{
    uint32_t addr;

    // ------------------------------------------------------------------------
    // Constructors 
    // ------------------------------------------------------------------------
    
    ipaddr();
    ipaddr(uint32_t);
    ipaddr(const ipaddr&);
    ipaddr(const char*);
    ipaddr(const in_addr&);
    ipaddr(const uint8_t*);
    ipaddr(const sockaddr&);
    ipaddr(const std::string&);

    // ------------------------------------------------------------------------
    //  Byte Ordering
    // ------------------------------------------------------------------------
    
    void change_byte_ordering();

    // ------------------------------------------------------------------------
    // String Representation
    // ------------------------------------------------------------------------

    void  fill_string(char* in) const;
    void  fill_string(std::string& in) const;
    char* toString() const;

    // ------------------------------------------------------------------------
    // Casting Operators 
    // ------------------------------------------------------------------------

    operator char* () const;
    operator uint32_t () const;
    operator std::string() const;
    operator const char* () const;

    // ------------------------------------------------------------------------
    // Binary Operators 
    // ------------------------------------------------------------------------

    bool   operator ! ();
    ipaddr operator ~ ();
    ipaddr operator & (const ipaddr&);
    ipaddr operator & (uint32_t);
    ipaddr& operator &= (const ipaddr&);
    ipaddr& operator &= (uint32_t);
    ipaddr operator | (const ipaddr&);
    ipaddr operator | (uint32_t);
    ipaddr& operator |= (const ipaddr&);
    ipaddr& operator |= (uint32_t);

    // ------------------------------------------------------------------------
    // Mathematical operators
    // ------------------------------------------------------------------------

    ipaddr operator ++ ();
    ipaddr operator -- ();
    ipaddr operator += (int);
    ipaddr operator += (const ipaddr&);
    ipaddr operator + (int);
    ipaddr operator + (const ipaddr&);
    int operator -(const ipaddr &);


    // ------------------------------------------------------------------------
    // Assignment Operator
    // ------------------------------------------------------------------------

    ipaddr& operator = (const ipaddr&);
    ipaddr& operator = (const std::string &);
    ipaddr& operator = (uint32_t);

    // ------------------------------------------------------------------------
    // Comparison Operators
    // ------------------------------------------------------------------------


    bool operator == (const ipaddr&);
    bool operator == (uint32_t);
    bool operator != (const ipaddr&);
    bool operator != (uint32_t);
    bool operator <  (uint32_t);
    bool operator <  (const ipaddr&);
    bool operator <= (uint32_t);
    bool operator <= (const ipaddr&);
    bool operator >  (uint32_t);
    bool operator >  (const ipaddr&);
    bool operator >= (uint32_t);
    bool operator >= (const ipaddr&);


}; // -- struct ipaddr

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr()
{
    addr = 0;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const char* addr_in)
{
    struct hostent* hp = 0;

    // -- REQUIRES
    assert(addr_in);

    if ((hp = ::gethostbyname(addr_in)) == 0)
    {
        perror("gethostbyname");
        throw std::runtime_error("could not convert string to address");
    }

    // -- CHECK
    assert(hp->h_length == 4);

    memcpy(&addr,hp->h_addr,hp->h_length);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const ipaddr& in)
{
    addr = in.addr;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const std::string& addr_in)
{
    struct hostent* hp = 0;

    if ((hp = ::gethostbyname(addr_in.c_str())) == 0)
    {
        perror("gethostbyname");
        throw std::runtime_error("could not convert string to address");
    }

    // -- CHECK
    assert(hp->h_length == 4);

    memcpy(&addr,hp->h_addr,hp->h_length);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const sockaddr& addr_in)
{
    addr = *((uint32_t*)&addr_in);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const in_addr& addr_in)
{
    addr = *((uint32_t*)&addr_in);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(const uint8_t* addr_in)
{
    // -- REQUIRES
    assert(addr_in);

    addr = *((uint32_t*)&addr_in);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::ipaddr(uint32_t addr_in)
{
    addr = htonl(addr_in);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
void 
ipaddr::change_byte_ordering()
{
    addr = htonl(addr);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
void 
ipaddr::fill_string(std::string& in) const
{
    char* tmp = 0;

    if((tmp = ::inet_ntoa(*((struct in_addr*)&addr))) == 0)
    {
        perror("inet_ntoa");
        throw std::runtime_error("could not convert address to string");
    }

    in = tmp;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
void 
ipaddr::fill_string(char* in) const
{
    // -- REQUIRES
    assert(in);

    char* tmp = 0;

    if((tmp = ::inet_ntoa(*((struct in_addr*)&addr))) == 0)
    {
        perror("inet_ntoa");
        throw std::runtime_error("could not convert address to string");
    }

    strncpy(in,tmp,16);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
char* 
ipaddr::toString() const
{ 
    static char ip_buf[16]; 
    const char* tmp = 0;

    if((tmp = ::inet_ntoa(*((struct in_addr*)&addr))) == 0)
    {
        perror("inet_ntoa");
        throw std::runtime_error("could not convert address to string");
    }

    strncpy(ip_buf,tmp,16);

    return ip_buf;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::operator std::string() const
{
    char* tmp = 0;

    if((tmp = ::inet_ntoa(*((struct in_addr*)&addr))) == 0)
    {
        perror("inet_ntoa");
        throw std::runtime_error("could not convert address to string");
    }

    return std::string(tmp);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::operator const char* () const
{ 
    return (const char*)toString();
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::operator uint32_t () const
{ 
    return addr; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr::operator char* () const
{ 
    return toString();
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator ~ ()
{
    return ipaddr(htonl(~addr));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator & (const ipaddr& in)
{
    return ipaddr(htonl(addr&in.addr));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator & (uint32_t in)
{
    return ipaddr(htonl(addr & in));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator &= (const ipaddr& in)
{
    addr &= in.addr;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator &= (uint32_t in)
{
    addr &= in;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator | (const ipaddr& in)
{
    return ipaddr(htonl(addr|in.addr));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator | (uint32_t in)
{
    return ipaddr(htonl(addr | in));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator |= (const ipaddr& in)
{
    addr |= in.addr;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator |= (uint32_t in)
{
    addr |= in;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator = (const std::string& in)
{
    struct hostent* hp = 0;

    if ((hp = ::gethostbyname(in.c_str())) == 0)
    {
        perror("gethostbyname");
        throw std::runtime_error("could not convert string to address");
    }

    // -- CHECK
    assert(hp->h_length == 4);

    memcpy(&addr,hp->h_addr,hp->h_length);
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator = (const ipaddr& in)
{
    addr = in.addr;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr& 
ipaddr::operator = (uint32_t in)
{
    addr = in;
    return *this; 
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator == (const ipaddr& in)
{
    return addr == in.addr;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator == (uint32_t in)
{
    return addr == in;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator != (const ipaddr& in)
{
    return addr != in.addr;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator != (uint32_t in)
{
    return addr != in;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator ++ ()
{
    addr = htonl(htonl(addr) + 1); 
    return *this;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator -- ()
{
    addr = htonl(htonl(addr) - 1); 
    return *this;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator += (int in)
{
    addr = htonl(htonl(addr) + in);
    return *this;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator += (const ipaddr& in)
{
    addr = htonl(htonl(addr) + htonl(in.addr));
    return *this;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator + (int in)
{
    return ipaddr(htonl(addr) + in);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
ipaddr 
ipaddr::operator + (const ipaddr& in)
{
    return ipaddr(htonl(addr) + htonl(in.addr));
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
int
ipaddr::operator -(const ipaddr &in)
{
    return htonl(addr) - htonl(in.addr);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator < (uint32_t in)
{
    return htonl(addr) < in;
}

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator < (const ipaddr& in)
{
    return htonl(addr) < htonl(in.addr);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator <= (uint32_t in)
{
    return htonl(addr) <= in;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator <= (const ipaddr& in)
{
    return htonl(addr) <= htonl(in.addr);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator >  (uint32_t in)
{
    return htonl(addr) > in;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator >  (const ipaddr& in)
{
    return htonl(addr) > htonl(in.addr);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator >= (uint32_t in)
{
    return htonl(addr) >= in;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool   
ipaddr::operator ! ()
{
    return addr == 0x00000000;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
inline
bool 
ipaddr::operator >= (const ipaddr& in)
{
    return htonl(addr) >= htonl(in.addr);
}
//-----------------------------------------------------------------------------

} // -- namespace pacman

#endif  // -- ipaddr_HH
