//
// yami++.h
//
// declaration of all the YAMI C++ wrappers
//
// Copyright (C) 2001-2003 Maciej Sobczak,
// student at the Institute of Computer Science,
// Warsaw University of Technology
//
// Permission to copy, use, modify, sell and distribute this software
// is granted provided this copyright notice appears in all copies.
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//

#ifndef YAMIPP_INCLUDED
#define YAMIPP_INCLUDED

#include <stdexcept>
#include <string>

// use alternative std::auto_ptr if there is no
// correct implementation in the target environment
#ifndef YAMI_NO_AUTO_PTR
#include <memory>
#else
#include "autoptr.hpp"
#endif

namespace YAMI
{

//
// exceptions
//

class LogicException : public std::logic_error
{
public:
     explicit LogicException(int cc);
     int getErrorCode() const;
private:
     int cc_;
};

class RunTimeException : public std::runtime_error
{
public:
     explicit RunTimeException(int cc);
     int getErrorCode() const;
private:
     int cc_;
};

class BadAddress : public RunTimeException
{
public:
     explicit BadAddress(int cc);
};

class Overflow : public RunTimeException
{
public:
     explicit Overflow();
};

class OSError : public RunTimeException
{
public:
     explicit OSError(int cc);
};

class InternalException : public RunTimeException
{
public:
     explicit InternalException();
};

// forwards:
class Agent;
class PoliciesImpl;
class Message;
class IncomingMsg;

//
// parameter set
//

// note:
// This class is not intended to be a base class.
class ParamSet
{
public:
     enum eType { eNoType = 0, eString = 1, eWString = 2, eInt = 3,
          eDouble = 4, eByte = 5, eBinary = 6 };
     
     // constructor/destructor
     ParamSet(int parcount);
     ParamSet(void *handle, bool owner = true); // throw()
     ~ParamSet(); // throw()
     
     // copy
     ParamSet(const ParamSet &ps);
     ParamSet & operator=(const ParamSet &ps);
     
     // copy of one element
     void copyParameter(int positiondst, const ParamSet &ps, int positionsrc);
     
     // setters

     void setInt(int position, int value);
     void setByte(int position, char value);
     void setDouble(int position, double value);
     void setBinary(int position, const char *buf, int bufsize);
     void setBinaryShallow(int position, const char *buf, int bufsize);
     void setString(int position, const char *str);
     void setString(int position, const std::string &str);
     void setStringShallow(int position, const char *str);
     void setWString(int position, const wchar_t *wstr);
#ifndef YAMI_NO_WSTRING
     void setWString(int position, const std::wstring &wstr);
#endif
     void setWStringShallow(int position, const wchar_t *wstr);

     // getters
     
     int getParamCount() const;
     eType getType(int position) const;
     bool isASCII(int position) const;

     int             getInt(int position) const;
     char            getByte(int position) const;
     double          getDouble(int position) const;
     int             getBinarySize(int position) const;
     void            getBinaryValue(int position, char *buf) const;
     const char *    getBinaryBuffer(int position) const;
     int             getStringLength(int position) const;
     void            getStringValue(int position, char *buf) const;
     const char *    getStringBuffer(int position) const;
     void            getString(int position, std::string &str) const;
     int             getWStringLength(int position) const;
     void            getWStringValue(int position, wchar_t *buf) const;
     const wchar_t * getWStringBuffer(int position) const;
#ifndef YAMI_NO_WSTRING
     void            getWString(int position, std::wstring &wstr) const;
#endif
     
     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(ParamSet &ps); // throw()

private:
     // isolated business -- in fact, the raw handle
     void* h_;
     bool owner_;
};

//
// agent
//

// incoming message wrapper
// note:
// This class is not intended to be a base class.
class IncomingMsg
{
public:
     IncomingMsg(void *handle, bool owner = true); // throw()
     ~IncomingMsg(); // throw()
     
     std::string getMsgName() const;
     std::string getObjectName() const;
     std::auto_ptr<ParamSet> getParameters() const;
     int getLevel() const;

     void eat();
     void reject();
     void reply();
     void reply(const ParamSet &paramset);

     void forward(const char *domainname, const char *objectname,
          const char *messagename);
     void forward(const std::string &domainname,
          const std::string &objectname, const std::string &messagename);
     void forward(const char *domainname, const char *objectname,
          const char *messagename, const ParamSet &paramset);
     void forward(const std::string &domainname,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     void forwardAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename);
     void forwardAddr(const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename);
     void forwardAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename,
          const ParamSet &paramset);
     void forwardAddr(const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(IncomingMsg &inc); // throw()

private:
     // copyright
     IncomingMsg(const IncomingMsg&);
     IncomingMsg& operator=(const IncomingMsg&);
     
     void *h_;
     bool owner_;
};

// interface of the passive object
// note:
// This class is intended TO BE a base class ;-)
class PassiveObject
{
public:
     virtual void call(IncomingMsg &msg) = 0;
};

// generic name (any object)
extern const char *AnyObject;

// name of the same domain (this domain)
extern const char *ThisDomain;

// message sent wrapper
// note:
// This class is not intended to be a base class.
class Message
{
public:
     enum eStatus { ePosted, ePending, eReplied, eRejected, eUnknownObj,
          eOverflow, eNetError, eTimedOut, eRejectByAgent };

     Message(void *handle, bool owner = true); // throw()
     ~Message(); // throw()
     
     void setTimeOut(int timeout);
     eStatus getStatus() const;
     void wait() const;
     std::auto_ptr<ParamSet> getResponse() const;
     
     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(Message &msg); // throw()

private:     
     // copyright
     Message(const Message&);
     Message& operator=(const Message&);

     void *h_;
     bool owner_;
};

// policies wrapper
// note:
// This class is not intended to be a base class.
class Policies
{
public:
     Policies(); // throw()
     ~Policies(); // throw()
     
     // setters // all throw()

     void setAgentLevel   (int  value); // 2
     void setObjQMaxLength(int  value); // 10
     void setObjQMaxSize  (int  value); // 1048576
     void setDispatchers  (int  value); // 1
     void setHasSender    (bool value); // true
     void setSQMaxLength  (int  value); // 256
     void setSQMaxSize    (int  value); // 1048576
     void setConnPoolSize (int  value); // 10
     void setSendTries    (int  value); // 5
     void setHasSocket    (bool value); // true
     void setReuseAddr    (bool value); // false
     void setHasReceiver  (bool value); // true
     void setHasWaker     (bool value); // true
     void setMTUser       (bool value); // true
     
private:     
     // copyright
     Policies(const Policies &);
     Policies& operator=(const Policies &);
     
     PoliciesImpl *pimpl_;
     
     friend class Agent;
};

// agent wrapper
// note:
// This class is not intended to be a base class.
class Agent
{
public:
     // helper enum for level description
     enum eAgentLevel { eLevel1 = 1, eLevel2 = 2 };
     
     // constructor/destructor
     Agent(int port);
     Agent(int port, const Policies &policies);
     Agent(void *handle, bool owner = true); // throw()
     ~Agent(); // throw()
     
     // domain set operations

     void domainRegister(const char *name, const char *addr,
          int port, int level);
     void domainRegister(const std::string &name, const std::string &addr,
          int port, int level);
     void domainUnregister(const char *name);
     void domainUnregister(const std::string &name);
     
     // object operation

     enum eObjectType { ePolling, ePassiveSingleThreaded,
                         ePassiveMultiThreaded };
     
     void objectRegister(const char *name, enum eObjectType type,
          PassiveObject *object);
     
     void objectRegister(const std::string &name, enum eObjectType type,
          PassiveObject *object);
     
     void objectUnregister(const char *name);
     void objectUnregister(const std::string &name);
     
     // incoming message operations

     std::auto_ptr<IncomingMsg> getIncoming(const char *objectname, bool wait);
     std::auto_ptr<IncomingMsg> getIncoming(const std::string &objectname,
          bool wait);

     // outgoing message operations

     std::auto_ptr<Message> send(const char *domainname,
          const char *objectname, const char *messagename);
     std::auto_ptr<Message> send(const std::string &domainname,
          const std::string &objectname, const std::string &messagename);
     std::auto_ptr<Message> send(const char *domainname,
          const char *objectname, const char *messagename,
          const ParamSet &paramset);
     std::auto_ptr<Message> send(const std::string &domainname,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     std::auto_ptr<Message> sendAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename);
     std::auto_ptr<Message> sendAddr(
          const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename);
     std::auto_ptr<Message> sendAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename,
          const ParamSet &paramset);
     std::auto_ptr<Message> sendAddr(
          const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     void sendOneWay(const char *domainname,
          const char *objectname, const char *messagename);
     void sendOneWay(const std::string &domainname,
          const std::string &objectname, const std::string &messagename);
     void sendOneWay(const char *domainname,
          const char *objectname, const char *messagename,
          const ParamSet &paramset);
     void sendOneWay(const std::string &domainname,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     void sendOneWayAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename);
     void sendOneWayAddr(const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename);
     void sendOneWayAddr(const char *addr, int port, int level,
          const char *objectname, const char *messagename,
          const ParamSet &paramset);
     void sendOneWayAddr(const std::string &addr, int port, int level,
          const std::string &objectname, const std::string &messagename,
          const ParamSet &paramset);
     
     // additional functions
     
     std::string getLocalAddress() const;
     bool processEvent(int timeout);
     
     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(Agent &a); // throw()

private:
     // copyright
     Agent(const Agent &);
     Agent& operator=(const Agent &);
     
     // isolated business
     void *h_;
     bool owner_;
};

// additional functions

void easySend(const char *address, int port, int level,
     const char *objectname, const char *messagename);
void easySend(const std::string &address, int port, int level,
     const std::string &objectname, const std::string &messagename);
void easySend(const char *address, int port, int level,
     const char *objectname, const char *messagename, const ParamSet &ps);
void easySend(const std::string &address, int port, int level,
     const std::string &objectname, const std::string &messagename,
     const ParamSet &ps);

// synchronization helpers

void sleep(int timeout);

class Mutex
{
public:
     Mutex();
     Mutex(void *handle, bool owner = true); // throw()
     ~Mutex(); // throw()
     
     void lock();
     void unlock();

     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(Mutex &m); // throw()

private:
     // copyright
     Mutex(const Mutex &);
     Mutex& operator=(const Mutex &);
     
     void *h_;
     bool owner_;
};

class MutexHolder
{
public:
     MutexHolder(Mutex &m) : m_(m)
     {
          m_.lock();
     }
     ~MutexHolder()
     {
          m_.unlock();
     }
     
private:
     // copyright
     MutexHolder(const MutexHolder &);
     MutexHolder& operator=(const MutexHolder &);

     Mutex &m_;
};

class Semaphore
{
public:
     Semaphore(int initval);
     Semaphore(void *handle, bool owner = true); // throw()
     ~Semaphore(); // throw()
     
     void acquire();
     bool tryAcquire();
     void release();

     // handle operations

     void * getHandle() const; // throw()
     void * resetHandle(void *handle, bool owner = true); // throw()
     void swap(Semaphore &s); // throw()

private:
     // copyright
     Semaphore(const Semaphore&);
     Semaphore& operator=(const Semaphore &);
     
     void *h_;
     bool owner_;
};

// additional functions

void netInitialize(void);
void netCleanup(void);

} // namespace YAMI

#endif /* YAMIPP_INCLUDED */
