/*
 * yamioslayer.c
 *
 * this is an abstract of the operating system, so that
 * the rest of the code will compile on most platforms
 *
 * Copyright (C) 2001-2003 Maciej Sobczak,
 * student at the Institute of Computer Science,
 * Warsaw University of Technology
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#include "../../include/yamic.h"
#include "../../include/yamierrors.h"
#include "yamioslayer.h"
#include "yamiXDR.h"
#include "yamitypes.h"
#include <string.h>

#ifndef WIN32
#include <unistd.h>

#ifndef YAMI_NO_THREADS
#include <pthread.h>
#endif

#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <errno.h>
#include <stdlib.h>
#else
#include <limits.h>
#include <Windows.h>
#include <Winsock.h>
#endif

#ifndef WIN32
#define closesocket close
#define SOCKET int
#define INVALID_SOCKET (-1)
#define SOCKET_ERROR (-1)
#endif


/* the following is for debugging */
/*#define YAMI_VERBOSE_MODE
*/

#ifdef YAMI_VERBOSE_MODE
#include <stdio.h>
#define VERBOSE
#define VERBOSESTREAM stderr
#endif

/* this is for tracking memory leaks */
#ifdef YAMI_TRACK_MEMORY

#include <stdio.h>

#define MALLOC(p, s) { \
     p = malloc(s); \
     fprintf(stderr, "%s(%d) malloc %p\n", __FILE__, __LINE__, p); \
}
#define FREE(a) { \
     fprintf(stderr, "%s(%d) free %p\n", __FILE__, __LINE__, a); \
     free(a); \
}

#else

#define MALLOC(p, s) { p = malloc(s); }
#define FREE(a) free(a);

#endif


/*
 * additional synchronization stuff
 */

#ifndef YAMI_NO_THREADS
struct sflag
{
     enum eYamiType rtti;
#ifndef WIN32
     pthread_mutex_t  mutex;
#else
     CRITICAL_SECTION critsec;
#endif
     int value;
};

int _checkFLAG(HFLAG f)
{
     if (f == NULL)
          return YAMIBADHANDLE;
     if (((struct sflag*)f)->rtti != eFlag)
          return YAMIBADHANDLE;
     return YAMIOK;
}
#endif /* YAMI_NO_THREADS */

int yamiCreateFlag(HFLAG *pf, int initval)
{
#ifdef YAMI_NO_THREADS
     /* no threads -> no flags */
     return YAMINOSUCHSERVICE;
#else

     struct sflag *pflag;
#ifndef WIN32
     int cc;
#endif

     MALLOC(pflag, sizeof(struct sflag))
     
     if (pflag == NULL)
          return YAMINOMEMORY;

     pflag->rtti = eFlag;
          
#ifndef WIN32
     cc = pthread_mutex_init(&(pflag->mutex), NULL);
     if (cc != 0)
     {
          FREE(pflag)
          return YAMISYNCHROERROR;
     }
#else
     InitializeCriticalSection(&(pflag->critsec));
#endif

     pflag->value = initval;

     *pf = pflag;
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiDestroyFlag(HFLAG f)
{
#ifdef YAMI_NO_THREADS
     /* no threads -> no flags */
     return YAMINOSUCHSERVICE;
#else

     struct sflag *pflag;
     int cc;

     if ((cc = _checkFLAG(f)) != YAMIOK)
          return cc;

     pflag = (struct sflag*)f;

#ifndef WIN32
     cc = pthread_mutex_destroy(&(pflag->mutex));
     if (cc != 0)
     {
          FREE(f)
          return YAMISYNCHROERROR;
     }
#else
     DeleteCriticalSection(&(pflag->critsec));
#endif
     
     FREE(f)
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiFlagSet(HFLAG f, int value)
{
#ifdef YAMI_NO_THREADS
     /* no threads -> no flags */
     return YAMINOSUCHSERVICE;
#else

     struct sflag *pflag;
     int cc;

     if ((cc = _checkFLAG(f)) != YAMIOK)
          return cc;

     pflag = (struct sflag*)f;

#ifndef WIN32
     cc = pthread_mutex_lock(&(pflag->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     EnterCriticalSection(&(pflag->critsec));
#endif

     pflag->value = value;

#ifndef WIN32
     cc = pthread_mutex_unlock(&(pflag->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     LeaveCriticalSection(&(pflag->critsec));
#endif     
     
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiFlagGet(HFLAG f, int *pvalue)
{
#ifdef YAMI_NO_THREADS
     /* no threads -> no flags */
     return YAMINOSUCHSERVICE;
#else

     struct sflag *pflag;
     int cc;

     if ((cc = _checkFLAG(f)) != YAMIOK)
          return cc;

     pflag = (struct sflag*)f;

#ifndef WIN32
     cc = pthread_mutex_lock(&(pflag->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     EnterCriticalSection(&(pflag->critsec));
#endif

     *pvalue = pflag->value;

#ifndef WIN32
     cc = pthread_mutex_unlock(&(pflag->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     LeaveCriticalSection(&(pflag->critsec));
#endif     
     
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

/*
 * thread-related abstracts
 */

#ifndef YAMI_NO_THREADS
struct sthread
{
     enum eYamiType rtti;
#ifndef WIN32
     pthread_t      tid;
#else
     HANDLE         tid;
#endif
     HFLAG stopflag;
};

int _checkTHREAD(HTHREAD th)
{
     if (th == NULL)
          return YAMIBADHANDLE;
     if (((struct sthread*)th)->rtti != eThread)
          return YAMIBADHANDLE;
     return YAMIOK;
}

struct threadparam
{
     THREADFUNCTION tf;
     HYAMIAGENT     agent;
     HSEMAPHORE     start_sem;
     HFLAG          stopflag;
};

#ifndef WIN32
void * _yamiThreadStarter(void *p)
#else
DWORD WINAPI _yamiThreadStarter(void *p)
#endif
{
     struct threadparam *param;
     THREADFUNCTION tf;
     HYAMIAGENT agent;
     HSEMAPHORE start_sem;
     HFLAG stopflag;

     param = (struct threadparam*)p;
     
     tf = param->tf;
     agent = param->agent;
     start_sem = param->start_sem;
     stopflag = param->stopflag;
     
     FREE(p)
     
     /* execute the actual thread function */
     tf(agent, start_sem, stopflag);
     
#ifndef WIN32
     return NULL;
#else
     return 0;
#endif
}
#endif /* YAMI_NO_THREADS */

int yamiStartNewThread(HTHREAD *pt, THREADFUNCTION tf, HYAMIAGENT agent)
{
#ifdef YAMI_NO_THREADS
     /* no threads */
     return YAMINOSUCHSERVICE;
#else

     struct threadparam *param;
     struct sthread *th;
     HSEMAPHORE start_sem;
     HFLAG stopflag;
#ifdef WIN32
     HANDLE thread;
     DWORD threadid;
#endif
     int cc;
     
     MALLOC(th, sizeof(struct sthread))
     
     if (th == NULL)
          return YAMINOMEMORY;

     th->rtti = eThread;
          
     MALLOC(param, sizeof(struct threadparam))
     
     if (param == NULL)
     {
          FREE(th)
          return YAMINOMEMORY;
     }
     
     cc = yamiCreateSemaphore(&start_sem, 0);
     if (cc != YAMIOK)
     {
          FREE(th)
          FREE(param)
          return YAMISYNCHROERROR;
     }

     cc = yamiCreateFlag(&stopflag, 0);
     if (cc != YAMIOK)
     {
          FREE(th)
          FREE(param)
          yamiDestroySemaphore(start_sem);
          return YAMISYNCHROERROR;
     }

     th->stopflag = stopflag;

     param->tf = tf;
     param->agent = agent;
     param->start_sem = start_sem;
     param->stopflag = th->stopflag;

#ifndef WIN32
     cc = pthread_create(&(th->tid), NULL, _yamiThreadStarter, param);
     if (cc != 0)
#else
     thread = CreateThread(NULL, 0, _yamiThreadStarter, param, 0, &threadid);
     if (thread == NULL)
#endif
     {
          FREE(th)
          FREE(param)
          yamiDestroySemaphore(start_sem);
          yamiDestroyFlag(stopflag);
          return YAMICANNOTSTARTTHREAD;
     }

#ifdef WIN32
     th->tid = thread;
#endif
     
     /* wait until the thread sets its started event */
     cc = yamiSemaphoreAcquire(start_sem);
     if (cc != YAMIOK)
     {
          FREE(th)
          return YAMISYNCHROERROR;
     }

     cc = yamiDestroySemaphore(start_sem);
     if (cc != YAMIOK)
     {
          FREE(th)
          return cc;
     }

     *pt = th;
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

#ifndef YAMI_NO_THREADS
int _checkSEMAPHORE(HSEMAPHORE s);
#endif /* YAMI_NO_THREADS */

int yamiStopThread(HTHREAD t, HSEMAPHORE sem)
{
#ifdef YAMI_NO_THREADS
     if (t == NULL)
          return YAMIOK;

     /* this should never be called, because there is no way    */
     /* to have t != NULL with this flag set during compilation */
     return YAMINOSUCHSERVICE;
#else

     struct sthread *th;
#ifdef WIN32
     DWORD wait;
#endif
     int cc;

     if (t == NULL)
          return YAMIOK;
     if ((cc = _checkTHREAD(t)) != YAMIOK)
          return cc;
     
     th = (struct sthread*)t;
     
     /* set the stop event */
     cc = yamiFlagSet(th->stopflag, 1);
     if (cc != YAMIOK)
     {
          return cc;
     }
     
     /* if the semaphore handle is valid, release it */
     if ((cc = _checkSEMAPHORE(sem)) == YAMIOK)
     {
          cc = yamiSemaphoreRelease(sem);
          if (cc != YAMIOK)
          {
               return cc;
          }
     }
     
     /* wait for a thread to terminate */
#ifndef WIN32
     cc = pthread_join(th->tid, NULL);
     if (cc != 0)
          return YAMITHREADERROR;
#else
     wait = WaitForSingleObject(th->tid, INFINITE);
     if (wait != WAIT_OBJECT_0)
          return YAMITHREADERROR;

     CloseHandle(th->tid);
#endif

     cc = yamiDestroyFlag(th->stopflag);
     if (cc != YAMIOK)
     {
          FREE(th)
          return cc;
     }

     FREE(th)
     
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiStopManyThreads(HTHREAD *t, HSEMAPHORE sem, int number)
{
#ifdef YAMI_NO_THREADS
     if (t == NULL)
          return YAMIOK;

     /* this should never be called - there is no way to have */
     /* t != NULL with this flag set during compilation       */
     return YAMINOSUCHSERVICE;
#else

     struct sthread *th;
#ifdef WIN32
     DWORD wait;
#endif
     int cc, i;

     if (t == NULL)
          return YAMIOK;
     if ((cc = _checkSEMAPHORE(sem)) != YAMIOK)
          return cc;

     /* first, set stop events for all the threads */
     for (i = 0; i < number; ++i)
     {
          if ((cc = _checkTHREAD(t[i])) != YAMIOK)
               return cc;

          th = (struct sthread*)(t[i]);

          /* set the stop event */
          cc = yamiFlagSet(th->stopflag, 1);
          if (cc != YAMIOK)
               return cc;
     }

     /* now, release the semaphore for every thread */
     for (i = 0; i < number; ++i)
     {
          cc = yamiSemaphoreRelease(sem);
          if (cc != YAMIOK)
          {
               return cc;
          }
     }
     
     /* and wait for every thread to terminate */
     for (i = 0; i < number; ++i)
     {
          th = (struct sthread*)(t[i]);

#ifndef WIN32
          cc = pthread_join(th->tid, NULL);
          if (cc != 0)
               return YAMITHREADERROR;

#else
          wait = WaitForSingleObject(th->tid, INFINITE);
          if (wait != WAIT_OBJECT_0)
               return YAMITHREADERROR;

          CloseHandle(th->tid);
#endif

          cc = yamiDestroyFlag(th->stopflag);
          if (cc != YAMIOK)
          {
               FREE(th)
               return cc;
          }

          FREE(th)
     }

     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiStopReceiverThread(HTHREAD t, int ipaddr, int ipport)
{
#ifdef YAMI_NO_THREADS
     if (t == NULL)
          return YAMIOK;     

     /* this shoul never be called - there is no way to have t != NULL */
     /* with this flag set during compilation                          */
     return YAMINOSUCHSERVICE;
#else

     struct sthread *th;
     struct sockaddr_in servaddr;
     SOCKET sd;
#ifdef WIN32
     DWORD wait;
#endif
     int cc;

     if (t == NULL)
          return YAMIOK;     
     if ((cc = _checkTHREAD(t)) != YAMIOK)
          return cc;
     
     th = (struct sthread*)t;
     
     /* set the stop event */
     cc = yamiFlagSet(th->stopflag, 1);
     if (cc != YAMIOK)
          return cc;

     /* send a dummy packet to the listening socket */
     /* it will wake up the receiver                */
     sd = socket(AF_INET, SOCK_STREAM, 0);
#ifndef WIN32
     if (sd < 0)
#else
     if (sd == INVALID_SOCKET)
#endif
          return YAMINETERROR;

     memset(&servaddr, 0, sizeof(servaddr));
     servaddr.sin_family = AF_INET;
#ifndef WIN32
     servaddr.sin_port = htons(ipport);
#else
     servaddr.sin_port = htons((u_short)ipport);
#endif
     memcpy(&(servaddr.sin_addr), &ipaddr, sizeof(int));

     if (connect(sd, (struct sockaddr*)&servaddr, sizeof(servaddr)) < 0)
     {
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: send error while stopping the "
               "receiver: port: %d, %s\n",
               ipport, strerror(errno));
#endif
          return YAMINETERROR;
     }
     closesocket(sd);

     /* wait for a thread to terminate */
#ifndef WIN32
     cc = pthread_join(th->tid, NULL);
     if (cc != 0)
          return YAMITHREADERROR;
#else
     wait = WaitForSingleObject(th->tid, INFINITE);
     if (wait != WAIT_OBJECT_0)
          return YAMITHREADERROR;

     CloseHandle(th->tid);
#endif

     cc = yamiDestroyFlag(th->stopflag);
     if (cc != YAMIOK)
     {
          FREE(th)
          return cc;
     }
     
     FREE(th)
     
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

#ifndef YAMI_NO_THREADS
struct smutex
{
     enum eYamiType   rtti;
#ifndef WIN32
     pthread_mutex_t  mutex;
#else
     CRITICAL_SECTION critsec;
#endif
};

int _checkMUTEX(HMUTEX m)
{
     if (m == NULL)
          return YAMIBADHANDLE;
     if (((struct smutex*)m)->rtti != eMutex)
          return YAMIBADHANDLE;
     return YAMIOK;
}
#endif /* YAMI_NO_THREADS */

int yamiCreateMutex(HMUTEX *pm)
{
#ifdef YAMI_NO_THREADS
     /* no threads */
     return YAMINOSUCHSERVICE;
#else

     struct smutex *pmutex;
#ifndef WIN32
     int cc;
#endif

     MALLOC(pmutex, sizeof(struct smutex))
     
     if (pmutex == NULL)
          return YAMINOMEMORY;

     pmutex->rtti = eMutex;
          
#ifndef WIN32
     cc = pthread_mutex_init(&(pmutex->mutex), NULL);
     if (cc != 0)
     {
          FREE(pmutex)
          return YAMISYNCHROERROR;
     }
#else
     InitializeCriticalSection(&(pmutex->critsec));
#endif

     *pm = pmutex;
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}


/* does not care about any waiting thread */
int yamiDestroyMutex(HMUTEX m)
{
#ifdef YAMI_NO_THREADS
     if (m == NULL)
          return YAMIOK;

     /* this should never happen - there is no way to have m != NULL */
     /* with this flag set during compilation                        */
     return YAMINOSUCHSERVICE;
#else

     struct smutex *pmutex;
     int cc;

     if (m == NULL)
          return YAMIOK;
     if ((cc = _checkMUTEX(m)) != YAMIOK)
          return cc;

     pmutex = (struct smutex*)m;

#ifndef WIN32
     cc = pthread_mutex_destroy(&(pmutex->mutex));
     if (cc != 0)
     {
          FREE(m)
          return YAMISYNCHROERROR;
     }
#else
     DeleteCriticalSection(&(pmutex->critsec));
#endif
     
     FREE(m)
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiMutexLock(HMUTEX m)
{
#ifdef YAMI_NO_THREADS
     if (m == NULL)
          return YAMIOK;

     /* this should never happen - there is no way to have m != NULL */
     /* with this flag set during compilation                        */
     return YAMINOSUCHSERVICE;
#else

     struct smutex *pmutex;
     int cc;

     if (m == NULL)
          return YAMIOK;
          
     if ((cc = _checkMUTEX(m)) != YAMIOK)
          return cc;
     
     pmutex = (struct smutex*)m;

#ifndef WIN32
     cc = pthread_mutex_lock(&(pmutex->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     EnterCriticalSection(&(pmutex->critsec));
#endif     

     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

int yamiMutexUnlock(HMUTEX m)
{
#ifdef YAMI_NO_THREADS
     if (m == NULL)
          return YAMIOK;     

     /* this should never happen - there is no way to have m != NULL */
     /* with this flag set during compilation                        */
     return YAMINOSUCHSERVICE;
#else

     struct smutex *pmutex;
     int cc;

     if (m == NULL)
          return YAMIOK;     

     if ((cc = _checkMUTEX(m)) != YAMIOK)
          return cc;
     
     pmutex = (struct smutex*)m;

#ifndef WIN32
     cc = pthread_mutex_unlock(&(pmutex->mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     LeaveCriticalSection(&(pmutex->critsec));
#endif     
     
     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

/* note: for YAMI_NO_THREADS, we make an exception with semaphores */
/* they can exist, but no blocking is possible with them,          */
/* which means that only yamiSemaphoreAcquire is forbidden         */
/* - then, semaphore works like a plain counter                    */
struct semaphore
{
     enum eYamiType  rtti;
#ifndef WIN32
     int             semval;

#ifndef YAMI_NO_THREADS
     pthread_cond_t  semval_cond;
     pthread_mutex_t semval_mutex;
#endif

#else
     HANDLE          semaphore;
#endif
};

int _checkSEMAPHORE(HSEMAPHORE s)
{
     if (s == NULL)
          return YAMIBADHANDLE;
     if (((struct semaphore*)s)->rtti != eSemaphore)
          return YAMIBADHANDLE;
     return YAMIOK;
}

int yamiCreateSemaphore(HSEMAPHORE *ps, int initval)
{
     struct semaphore *psemaphore;
#ifndef WIN32
     int cc;
#else
     HANDLE sem;
#endif

     MALLOC(psemaphore, sizeof(struct semaphore))
     
     if (psemaphore == NULL)
          return YAMINOMEMORY;

     psemaphore->rtti = eSemaphore;

#ifndef WIN32          
     psemaphore->semval = initval;

#ifndef YAMI_NO_THREADS
     cc = pthread_cond_init(&(psemaphore->semval_cond), NULL);
     if (cc != 0)
     {
          FREE(psemaphore)
          return YAMISYNCHROERROR;
     }
     cc = pthread_mutex_init(&(psemaphore->semval_mutex), NULL);
     if (cc != 0)
     {
          FREE(psemaphore)
          return YAMISYNCHROERROR;
     }
#endif /* YAMI_NO_THREADS */

#else
     sem = CreateSemaphore(NULL, initval, INT_MAX, NULL);
     if (sem == NULL)
     {
          FREE(psemaphore)
          return YAMISYNCHROERROR;
     }
     psemaphore->semaphore = sem;
#endif
     
     *ps = psemaphore;
     return YAMIOK;
}


/* this function doesn't care about the threads waiting on this semaphore */
int yamiDestroySemaphore(HSEMAPHORE s)
{
     struct semaphore *psem;
     int cc1;
#ifndef WIN32
     int cc2;
#endif

     if (s == NULL)
          return YAMIOK;
     if ((cc1 = _checkSEMAPHORE(s)) != YAMIOK)
          return cc1;

     psem = (struct semaphore*)s;

#ifndef WIN32

#ifndef YAMI_NO_THREADS
     cc1 = pthread_mutex_destroy(&(psem->semval_mutex));
     cc2 = pthread_cond_destroy(&(psem->semval_cond));
     if (cc1 != 0 || cc2 != 0)
     {
          FREE(s)
          return YAMISYNCHROERROR;
     }
#endif /* YAMI_NO_THREADS */

#else
     CloseHandle(psem->semaphore);
#endif

     FREE(s)
     return YAMIOK;
}

int yamiSemaphoreAcquire(HSEMAPHORE s)
{
#ifdef YAMI_NO_THREADS
     /* no threads -> acquire is forbidden, since it has blocking implied */
     return YAMINOSUCHSERVICE;
#else

     struct semaphore *psemaphore;
#ifdef WIN32
     DWORD wait;
#endif
     int cc;

     if ((cc = _checkSEMAPHORE(s)) != YAMIOK)
          return cc;

     psemaphore = (struct semaphore*)s;
     
#ifndef WIN32
     cc = pthread_mutex_lock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;

     while (psemaphore->semval == 0)
     {
          cc = pthread_cond_wait(&(psemaphore->semval_cond),
               &(psemaphore->semval_mutex));
          if (cc != 0)
          {
               pthread_mutex_unlock(&(psemaphore->semval_mutex));
               return YAMISYNCHROERROR;
          }
     }
     
     psemaphore->semval--;
     
     cc = pthread_mutex_unlock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#else
     wait = WaitForSingleObject(psemaphore->semaphore, INFINITE);
     if (wait != WAIT_OBJECT_0)
          return YAMISYNCHROERROR;
#endif

     return YAMIOK;

#endif /* YAMI_NO_THREADS */
}

/* note: TryAcquire is OK with YAMI_NO_THREADS flag */
int yamiSemaphoreTryAcquire(HSEMAPHORE s)
{
     struct semaphore *psemaphore;
#ifdef WIN32
     DWORD wait;
#else
     int done;
#endif
     int cc;
     
     if ((cc = _checkSEMAPHORE(s)) != YAMIOK)
          return cc;

     psemaphore = (struct semaphore*)s;
     
#ifndef WIN32

#ifndef YAMI_NO_THREADS
     cc = pthread_mutex_lock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#endif /* YAMI_NO_THREADS */

     if (psemaphore->semval > 0)
     {
          psemaphore->semval--;
          done = 1;
     }
     else
          done = 0;

#ifndef YAMI_NO_THREADS
     cc = pthread_mutex_unlock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#endif /* YAMI_NO_THREADS */
     
     return done ? YAMIOK : YAMINOTAVAILABLE;
#else
     wait = WaitForSingleObject(psemaphore->semaphore, 0);
     if (wait == WAIT_TIMEOUT)
          return YAMINOTAVAILABLE;
     if (wait == WAIT_OBJECT_0)
          return YAMIOK;
     else
          return YAMISYNCHROERROR;
#endif
}

/* note: Release is OK with YAMI_NO_THREADS flag */
int yamiSemaphoreRelease(HSEMAPHORE s)
{
     struct semaphore *psemaphore;
     int cc;

     if ((cc = _checkSEMAPHORE(s)) != YAMIOK)
          return cc;

     psemaphore = (struct semaphore*)s;

#ifndef WIN32     

#ifndef YAMI_NO_THREADS
     cc = pthread_mutex_lock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#endif /* YAMI_NO_THREADS */

     psemaphore->semval++;

#ifndef YAMI_NO_THREADS
     cc = pthread_cond_signal(&(psemaphore->semval_cond));
     if (cc != 0)
     {
          pthread_mutex_unlock(&(psemaphore->semval_mutex));
          return YAMISYNCHROERROR;
     }
     
     cc = pthread_mutex_unlock(&(psemaphore->semval_mutex));
     if (cc != 0)
          return YAMISYNCHROERROR;
#endif /* YAMI_NO_THREADS */

#else
     cc = ReleaseSemaphore(psemaphore->semaphore, 1, NULL);
     if (cc == 0)
          return YAMISYNCHROERROR;
#endif
     
     return YAMIOK;
}

int yamiSleep(int timeout)
{
#ifndef WIN32
     struct timeval tv;
     int rpt;
     int cc;
#else
     DWORD towait;
#endif
     
     if (timeout == 0)
     {
#ifndef WIN32
          timeout = 3600;
          rpt = 1;
#else
          towait = INFINITE;
#endif
     }
     else
     {
#ifndef WIN32
          rpt = 0;
#else
          towait = timeout;
#endif
     }

#ifndef WIN32
     do
     {     
          tv.tv_sec = timeout / 1000;
          tv.tv_usec = (timeout % 1000) * 1000;

          cc = select(0, NULL, NULL, NULL, &tv);
          if (cc != 0)
               return YAMISYNCHROERROR;
     } while (rpt);
#else
     Sleep(towait);
#endif
     
     return YAMIOK;
}


/*
 * network-related abstracts
 */

#ifdef WIN32
#define MAXHOSTNAMELEN 1024
#endif

int yamiNetGetLocalHost(int *ipaddr)
{
     char name[MAXHOSTNAMELEN];
     
     if (gethostname(name, MAXHOSTNAMELEN) != 0)
          return YAMINETERROR;
     
     return yamiNetResolveIPAddress(name, ipaddr);     
}

int yamiNetResolveIPAddress(const char *addr, int *ipaddr)
{
     struct hostent *he;

     if ((he = gethostbyname(addr)) == NULL)
          return YAMICANNOTRESOLVEADDRESS;

     if (he->h_length != sizeof(int))
          return YAMIBADADDRTYPE;

     *ipaddr = *(int*)(he->h_addr_list[0]);
     return YAMIOK;
}

/* one socket in a pool */
struct socketinpool
{
     int    addr;     /* address, network byte order */
     int    port;
     SOCKET sockd;     /* socket descriptor           */
};

/* implementation of the socket pool */
struct socketpool
{
     enum eYamiType      rtti;
     int                 poolsize;
     struct socketinpool pool[1];
};

int _checkSOCKETPOOL(HSOCKETPOOL sp)
{
     if (sp == NULL)
          return YAMIBADHANDLE;
     if (((struct socketpool*)sp)->rtti != eSocketPool)
          return YAMIBADHANDLE;
     return YAMIOK;
}

int yamiNetPoolCreate(HSOCKETPOOL *pool, int size)
{
     struct socketpool *p;
     size_t totalsize;

#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: creating the connection pool\n");
#endif

     totalsize = sizeof(enum eYamiType) +
               sizeof(int) +
               size * sizeof(struct socketinpool);
     MALLOC(p, totalsize)
     
     if (p == NULL)
          return YAMINOMEMORY;
     
     memset(p, 0, totalsize);
     
     p->rtti = eSocketPool;
     
     p->poolsize = size;

     *pool = p;
     return YAMIOK;
}

int yamiNetPoolDestroy(HSOCKETPOOL pool)
{
     struct socketpool *p;
     int cc, i;

     if (pool == NULL)
          return YAMIOK;
     if ((cc = _checkSOCKETPOOL(pool)) != YAMIOK)
          return cc;
          
#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: destroying the connection pool\n");
#endif

     p = (struct socketpool*)pool;
     
     for (i = 0; i < p->poolsize; ++i)
          if (p->pool[i].sockd)
               closesocket(p->pool[i].sockd);
     
     FREE(pool)
     
     return YAMIOK;
}

int _findempty(struct socketpool *p, int *pi)
{
     int i;

     for (i = 0; i < p->poolsize; ++i)
          if (p->pool[i].addr == 0)
               break;
     if (i == p->poolsize)
     {
          /* not found - close one of them */
          i = rand() % p->poolsize;
#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: closing one connection in a pool to "
          "free the place, close for %d\n", p->pool[i].port);
#endif
          if (closesocket(p->pool[i].sockd) == -1)
               return YAMINETERROR;
     }
     
     *pi = i;
     return YAMIOK;
}

int _clearoneinpool(struct socketpool *p, SOCKET sd)
{
     int i;
     
     for (i = 0; i < p->poolsize; ++i)
          if (p->pool[i].sockd == sd)
          {
               if (closesocket(sd) < 0)
               {
#ifdef VERBOSE
                    fprintf(VERBOSESTREAM, "n: close error: %s\n",
                         strerror(errno));
#endif
                    return YAMINETERROR;
               }

               p->pool[i].addr = 0;
               p->pool[i].port = 0;
               p->pool[i].sockd = 0;
               
               return YAMIOK;
          }
     
     return YAMINOTFOUND;
}

int _connecttopool(struct socketpool *p, int ipaddr, int port, SOCKET *psd)
{
     int i, cc;
     SOCKET sd;
     struct sockaddr_in servaddr;
     
#ifdef VERBOSE
     fprintf(VERBOSESTREAM,
          "n: creating new connection in a pool for port %d\n", port);
#endif

     if ((cc = _findempty(p, &i)) != YAMIOK)
          return cc;
     
     /* i points to the first empty socket in a pool */
     
     sd = socket(AF_INET, SOCK_STREAM, 0);
#ifndef WIN32
     if (sd < 0)
#else
     if (sd == INVALID_SOCKET)
#endif
          return YAMINETERROR;
     
     memset(&servaddr, 0, sizeof(servaddr));
     servaddr.sin_family = AF_INET;
#ifndef WIN32
     servaddr.sin_port = htons(port);
#else
     servaddr.sin_port = htons((u_short)port);
#endif
     memcpy(&(servaddr.sin_addr), &ipaddr, sizeof(int));

     if (connect(sd, (struct sockaddr*)&servaddr, sizeof(servaddr)) < 0)
     {
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: send error: port: %d, %s\n",
               port, strerror(errno));
#endif
          return YAMINETERROR;
     }

     p->pool[i].addr = ipaddr;
     p->pool[i].port = port;
     p->pool[i].sockd = sd;

     *psd = sd;
     return YAMIOK;     
}

int _writetorawsocket(SOCKET sd, const char *msg, size_t size)
{
#ifndef WIN32
     void (*sigpipefunc)(int);
     ssize_t nwritten;
#else
     int nwritten;
#endif
     fd_set rset;
     struct timeval tv;

#ifndef WIN32     
     /* here we have to block the SIGPIPE signal           */
     /* it will be sent to the process if there is a write */
     /* to the socket, whose reading end is close          */
     /* it can happen while reusing connections            */

     sigpipefunc = signal(SIGPIPE, SIG_IGN);
     if (sigpipefunc == SIG_ERR)
          return YAMINETERROR;
#endif
     
     /* second, agents use sockets ONLY in simplex mode    */
     /* it means, that we can test the socket for validity */
     /* by testing if it's ready for reading               */
     /* if yes, it means, that the reading end of the      */
     /* connection has closed (we received FIN)            */

     memset(&tv, 0, sizeof(tv));
     FD_ZERO(&rset);
     FD_SET(sd, &rset);
     if (select(sd + 1, &rset, NULL, NULL, &tv) != 0)
     {
#ifndef WIN32
          signal(SIGPIPE, sigpipefunc);
#endif
          
          /* this means that the connection has been closed */
          /* or some other error has occured                */
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: detected closed end\n");
#endif
          return YAMINETERROR;
     }

#ifndef WIN32     
     while (size > 0)
     {
          if ((nwritten = write(sd, msg, size)) <= 0)
          {
               if (errno == EINTR)
                    nwritten = 0;
               else
               {
                    signal(SIGPIPE, sigpipefunc);
                    return YAMINETERROR;
               }
          }
          
          size -= nwritten;
          msg += nwritten;
     }
     
     /* now, we can put the old signal handler back */
     if (signal(SIGPIPE, sigpipefunc) == SIG_ERR)
          return YAMINETERROR;
#else
     if ((nwritten = send(sd, msg, size, 0)) == SOCKET_ERROR)
          return YAMINETERROR;
#endif

     return YAMIOK;
}

int _readfromrawsocket(SOCKET sd, char *buf, size_t toread, size_t *bytesread)
{
     size_t nleft;
#ifndef WIN32
     ssize_t nread;
#else
     int nread;
#endif
     
     nleft = toread;
     while (nleft > 0)
     {
#ifndef WIN32
          if((nread = read(sd, buf, nleft)) < 0)
          {
               if (errno == EINTR)
                    nread = 0;
               else
               {
#ifdef VERBOSE
                    fprintf(VERBOSESTREAM, "n: read error: %s\n",
                         strerror(errno));
#endif
                    return YAMINETERROR;
               }
#else
          if((nread = recv(sd, buf, nleft, 0)) == SOCKET_ERROR)
          {
#ifdef VERBOSE
               fprintf(VERBOSESTREAM, "n: read error: %s\n", strerror(errno));
#endif
               return YAMINETERROR;
#endif
          }
          else if (nread == 0)
               break;
          
          nleft -= nread;
          buf += nread;
     }
     
     *bytesread = (size_t)(toread - nleft);
     return YAMIOK;
}

int _waitforhandshake(SOCKET sd)
{
     size_t readn;
     char dummy;
     int cc;
     
     cc = _readfromrawsocket(sd, &dummy, 1, &readn);
     if (cc != YAMIOK)
          return cc;

     if (readn == 1)
          return YAMIOK;
     else
          return YAMINOHANDSHAKE;
}

int yamiNetPoolSend(HSOCKETPOOL pool, int ipaddr, int port,
          const char *msg, size_t size, int tries)
{
     struct socketpool *p;
     SOCKET sd;
     int cc, i;
     
     if ((cc = _checkSOCKETPOOL(pool)) != YAMIOK)
          return cc;
          
     p = (struct socketpool*)pool;

#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: send called for port %d, tries %d\n",
          port, tries);
#endif
     
     /* find the socket for reuse or create a new one */
     for (i = 0; i < p->poolsize; ++i)
          if (p->pool[i].addr == ipaddr && p->pool[i].port == port)
               break;

     if (i == p->poolsize)
     {
          /* no connection for reuse has been found */
          /* create new connection in a pool        */
          cc = _connecttopool(p, ipaddr, port, &sd);
          if (cc != YAMIOK)
               return cc;
     }
     else
     {
          /* found - reuse it */
          sd = p->pool[i].sockd;
     }

     /* try sending until success */
     while (tries--)
     {
          cc = _writetorawsocket(sd, msg, size);
          if (cc == YAMIOK)
               cc = _waitforhandshake(sd);
     
          if (cc == YAMIOK)
               return YAMIOK;
          
          /* there is an error while reusing the connection */
          /* it's possible that the server has closed it    */
          /* try with a different one                       */
          
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: error while reusing the connection\n");
#endif

          cc = _clearoneinpool(p, sd);
          if (cc != YAMIOK)
               return cc;

          cc = _connecttopool(p, ipaddr, port, &sd);
          if (cc != YAMIOK)
               return cc;
     }
     
#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: send: giving up\n");
#endif
     return YAMINETERROR;
}

/* this is an abstract of a socket */
struct socketobject
{
     enum eYamiType rtti;
     SOCKET         socketd;
};

int _checkSOCKET(HSOCKET s)
{
     if (s == NULL)
          return YAMIBADHANDLE;
     if (((struct socketobject*)s)->rtti != eSocket)
          return YAMIBADHANDLE;
     return YAMIOK;
}

/* opens a listening socket on a given port          */
/* reusesocket is a flag for setting socket's option */
int yamiNetListen(int port, int reuseaddr, HSOCKET *ps)
{
     SOCKET sockd;
     struct sockaddr_in servaddr;
     struct socketobject *psock;
     
#ifndef WIN32
     if ((sockd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
#else
     if ((sockd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
#endif
          return YAMINETERROR;
     
     memset(&servaddr, 0, sizeof(servaddr));
     servaddr.sin_family = AF_INET;
     servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
#ifndef WIN32
     servaddr.sin_port = htons(port);
#else     
     servaddr.sin_port = htons((u_short)port);
#endif

#ifndef WIN32
     /* only for Unix-like systems                 */
     /* set SO_REUSEADDR if the policy requires it */
     if (reuseaddr)
     {
#ifdef VERBOSE
          fprintf(VERBOSESTREAM,
               "n: setting SO_REUSEADDR flag for listener\n");
#endif
          if (setsockopt(sockd, SOL_SOCKET, SO_REUSEADDR,
               &reuseaddr, sizeof(int)) < 0)
               return YAMINETERROR;
     }
#endif
     
     if (bind(sockd, (struct sockaddr*)&servaddr, sizeof(servaddr))
          == SOCKET_ERROR)
     {
          closesocket(sockd);
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: cannot bind: %s\n", strerror(errno));
#endif
          return YAMINETERROR;
     }
     
     if (listen(sockd, 5) == SOCKET_ERROR)
     {
          closesocket(sockd);
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: cannot listen: %s\n", strerror(errno));
#endif
          return YAMINETERROR;
     }
     
     MALLOC(psock, sizeof(struct socketobject))
     
     if (psock == NULL)
     {
          closesocket(sockd);
          return YAMINOMEMORY;
     }

     psock->rtti = eSocket;
          
     psock->socketd = sockd;
     
     *ps = psock;
     
#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: listening on the port: %d\n", port);
#endif
     
     return YAMIOK;
}

/* helper for reading from a socket */
int yamiNetReadN(HSOCKET s, char *buf, size_t toread, size_t *bytesread)
{
     SOCKET sd;
     int cc;

     if ((cc = _checkSOCKET(s)) != YAMIOK)
          return cc;
     
     sd = ((struct socketobject*)s)->socketd;

     return _readfromrawsocket(sd, buf, toread, bytesread);     
}

/* helper for handshaking */
int yamiNetShakeHand(HSOCKET s)
{
     struct socketobject *psock;
     char dummy;
     int cc;
     
     if ((cc = _checkSOCKET(s)) != YAMIOK)
          return cc;
     
     psock = (struct socketobject*)s;
     
     dummy = 0;
     cc = _writetorawsocket(psock->socketd, &dummy, 1);
     return cc;
}

/* accepts a connection from an open, listening socket */
/* or from a pool of connection, whatever comes first  */
int yamiNetPoolAccept(HSOCKET listens, HSOCKETPOOL pool,
          HSOCKET *accepteds, int timeout)
{
     struct socketobject *psock, *newsock;
     struct socketpool *p;
     struct timeval timeoutval;
     fd_set rset;
     SOCKET sd;
     int cc, i;
     int maxd = 0;
     
     if ((cc = _checkSOCKET(listens)) != YAMIOK)
          return cc;
     
     if ((cc = _checkSOCKETPOOL(pool)) != YAMIOK)
          return cc;
     
     psock = (struct socketobject*)listens;
     p = (struct socketpool*)pool;
     
     /* wait for the pool or for the listening socket to be ready */
     FD_ZERO(&rset);
     FD_SET(psock->socketd, &rset);
#ifndef WIN32
     maxd = psock->socketd;
#endif
     for (i = 0; i < p->poolsize; ++i)
          if (p->pool[i].addr != 0)
          {
               sd = p->pool[i].sockd;
               FD_SET(sd, &rset);
#ifndef WIN32
               if (sd > maxd)
                    maxd = sd;
#endif
          }

     if (timeout >= 0)
     {
          /* do not block, but wait timeout millisecs */

          timeoutval.tv_sec = timeout / 1000;
          timeoutval.tv_usec = 1000 * (timeout % 1000);
          
          cc = select(maxd + 1, &rset, NULL, NULL, &timeoutval);
#ifndef WIN32
          if (cc < 0)
#else
          if (cc == SOCKET_ERROR)
#endif
               return YAMINETERROR;
          
          if (cc == 0)
               return YAMINOTAVAILABLE;
     }
     else
     {
          /* do blocking select */
          
          cc = select(maxd + 1, &rset, NULL, NULL, NULL);
#ifndef WIN32
          if (cc <= 0)
#else
          if (cc == SOCKET_ERROR)
#endif
               return YAMINETERROR;
     }

     MALLOC(newsock, sizeof(struct socketobject))
     
     if (newsock == NULL)
          return YAMINOMEMORY;

     newsock->rtti = eSocket;
     
     /* check first the listening socket */
     if (FD_ISSET(psock->socketd, &rset))
     {
          /* we got a new connection from the listening socket */

          /* find a place in a socket pool */
          if ((cc = _findempty(p, &i)) != YAMIOK)
          {
               FREE(newsock)
               return cc;
          }
     
          /* accept the connection */
#ifndef WIN32
          if ((sd = accept(psock->socketd, (struct sockaddr*)NULL, NULL)) < 0)
#else
          if ((sd = accept(psock->socketd, (struct sockaddr*)NULL, NULL))
               == INVALID_SOCKET)
#endif     
          {
#ifdef VERBOSE
               fprintf(VERBOSESTREAM, "n: accept error: %s\n",
                    strerror(errno));
#endif
               FREE(newsock)
               return YAMINETERROR;
          }

          p->pool[i].addr = 1;     /* doesn't matter, should be non-zero */
          p->pool[i].port = 1;
          p->pool[i].sockd = sd;
          
#ifdef VERBOSE
          fprintf(VERBOSESTREAM, "n: accepted new connection\n");
#endif
     }
     else
     {
          /* if not, there is some data to read */
          /* from the socket pool               */
          for (i = 0; i < p->poolsize; ++i)
          {
               sd = p->pool[i].sockd;
               if (FD_ISSET(sd, &rset))
                    break;
          }
     }
     
     newsock->socketd = sd;
     *accepteds =  newsock;
     return YAMIOK;
}

/* closes a socket */
int yamiNetCloseSocket(HSOCKET s)
{
     struct socketobject *psock;
     int cc;
     
     if ((cc = _checkSOCKET(s)) != YAMIOK)
          return cc;
     
     psock = (struct socketobject*)s;

     cc = closesocket(psock->socketd);
#ifdef VERBOSE
#ifndef WIN32
     if (cc != 0)
#else
     if (cc == SOCKET_ERROR)
#endif
          fprintf(VERBOSESTREAM, "n: close error: %s\n", strerror(errno));
#endif
     
     FREE(psock)
     
     return YAMIOK;     
}

int yamiNetDisposeButDontCloseSocket(HSOCKET s)
{
     int cc;
     
     if ((cc = _checkSOCKET(s)) != YAMIOK)
          return cc;
     
     FREE(s)
     return YAMIOK;     
}

int yamiNetPoolCloseOne(HSOCKETPOOL pool, HSOCKET toclose)
{
     struct socketobject* psock;
     struct socketpool *p;
     int cc;
     
     if ((cc = _checkSOCKET(toclose)) != YAMIOK)
          return cc;
     
     if ((cc = _checkSOCKETPOOL(pool)) != YAMIOK)
          return cc;
     
     psock = (struct socketobject*)toclose;
     p = (struct socketpool*)pool;

#ifdef VERBOSE
     fprintf(VERBOSESTREAM, "n: closing one connection in a pool\n");
#endif

     return _clearoneinpool(p, psock->socketd);
}

int yamiNetInitialize(void)
{
#ifndef WIN32
     return YAMIOK;
#else
     WORD versionRequested;
     WSADATA wsadata;
     int err;

     versionRequested = MAKEWORD(2, 0);

     err = WSAStartup(versionRequested, &wsadata);
     return (err == 0) ? YAMIOK : YAMINETERROR;
#endif
}

int yamiNetCleanup(void)
{
#ifndef WIN32
     return YAMIOK;
#else
     int err;

     err = WSACleanup();
     return (err == 0) ? YAMIOK : YAMINETERROR;
#endif
}
