/*
 * yamioslayer.h
 *
 * this is an abstract of the operating system, so that
 * the rest of the code will compile on most platforms
 *
 * Copyright (C) 2001-2003 Maciej Sobczak,
 * student at the Institute of Computer Science,
 * Warsaw University of Technology
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#ifndef YAMIOSLAYER_INCLUDED
#define YAMIOSLAYER_INCLUDED

#include "../../include/yamisynchro.h"
#include "../../include/yamic.h"
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * thread-related abstracts
 */

typedef void* HTHREAD;
typedef void* HFLAG;

/* start_sem is the handle of the semaphore that the thread       */
/* should push up as soon as it's ready to work                   */
/* stopevent is the address of the int variable                   */
/* if this variable is non-zero, thread function should terminate */
typedef void (*THREADFUNCTION)(HYAMIAGENT agent,
          HSEMAPHORE start_sem, HFLAG stopflag);

/* this function starts a new thread and returns a handle to it   */
int yamiStartNewThread(HTHREAD *pt, THREADFUNCTION tf, HYAMIAGENT agent);

/* this function sets the stopevent for a given thread, releases  */
/* the semaphore (if not NULL) and waits for it to terminate      */
int yamiStopThread(HTHREAD t, HSEMAPHORE sem);

/* this is special case                    */
/* stop all the threads in an array        */
int yamiStopManyThreads(HTHREAD *t, HSEMAPHORE sem, int number);

/* this is special case - or rather a hack */
/* stop the receiving thread               */
int yamiStopReceiverThread(HTHREAD t, int ipaddr, int ipprt);

/*
 * additional synchronization stuff
 */

/* create and destroy a flag object */
int yamiCreateFlag(HFLAG *pf, int initval);
int yamiDestroyFlag(HFLAG f);

/* set and test a flag */
int yamiFlagSet(HFLAG f, int value);
int yamiFlagGet(HFLAG f, int *pvalue);

/*
 * network-related abstracts
 */

typedef void* HSOCKET;
typedef void* HSOCKETPOOL;

/* get the ip address (in network-byte order) of the local host */
int yamiNetGetLocalHost(int *ipaddr);

/* this converts addr into ip address which is already in network-byte order */
int yamiNetResolveIPAddress(const char *addr, int *ipaddr);

/* creates an empty pool of network connections               */
int yamiNetPoolCreate(HSOCKETPOOL *pool, int size);

/* closes all cached connections and destroys the pool        */
int yamiNetPoolDestroy(HSOCKETPOOL pool);

/* this function connects to ipaddr, port and sends a message */
/* if the socket pool already contains the connection to the  */
/* given address, it's reused                                 */
/* if the pool is full and a new connection should be created */
/* one of the sockets in a pool gets closed                   */
/* the message is opaque: function does not interpret it      */
/* ipaddr is in network-byte order                            */
/* tries tells how many time this function should try to send */
/* the data before giving up with YAMINETERROR return code    */
int yamiNetPoolSend(HSOCKETPOOL pool, int ipaddr, int port,
          const char *msg, size_t size, int tries);

/* opens a listening socket on a given port          */
/* reusesocket is a flag for setting socket's option */
int yamiNetListen(int port, int reusesocket, HSOCKET *ps);

/* accepts a message from an open, listening socket           */
/* or from one of the sockets in a pool, whatever comes first */
/* if a new connection is accepted from the listening socket, */
/* it's automatically added to the pool                       */
/* the user should not close the accepted connection          */
/* (it is part of the pool)                                   */
/* the wait flag allows to wait for a connection or to quit   */
/* if there is none (YAMINOTAVAILABLE)                        */
int yamiNetPoolAccept(HSOCKET listens, HSOCKETPOOL pool,
          HSOCKET *accepteds, int timeout);

/* close one connection in a pool                   */
/* the connection closed is the one indicated       */
/* by the open socket, which is the copy of one     */
/* of the sockets held by the pool                  */
int yamiNetPoolCloseOne(HSOCKETPOOL pool, HSOCKET toclose);

/* helper for reading from a socket */
int yamiNetReadN(HSOCKET s, char *buf, size_t toread, size_t *bytesread);

/* helper for handshake */
int yamiNetShakeHand(HSOCKET s);

/* closes a socket */
int yamiNetCloseSocket(HSOCKET s);

/* releases the object, but does not close a socket */
int yamiNetDisposeButDontCloseSocket(HSOCKET s);

/* network initialization */
/* (for portability)      */
int yamiNetInitialize(void);

/* network cleanup   */
/* (for portability) */
int yamiNetCleanup(void);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
