//
// eserver.cc
//
// this file implements the Event Service, but the
// user is free to use it for any purpose s/he wants
//
// Copyright (C) 2001-2003 Maciej Sobczak,
// student at the Institute of Computer Science,
// Warsaw University of Technology
//
// Permission to copy, use, modify, sell and distribute this software
// is granted provided this copyright notice appears in all copies.
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//

#include "yami++.h"
#include <iostream>
#include <string>
#include <list>
#include <stdio.h> // for sprintf, sscanf

using namespace YAMI;


// this is the name of the event server's main object
const std::string es_main_objectname("eventserver");

// this is the name of the object used to close the server
// if the quit option is enabled
const std::string es_quit_objectname("eventserver_quit");


// the object of this class encapsulates the data
// stored for each listener
// the event server delegates to this object the actual
// task of relaying the message
class Subscriber
{
public:

     // the constructor is responsible for registering the new
     // event listener in the local address book and for
     // replying to the subscriber with the new subscription id
     Subscriber(const std::string &address, int port, int level,
          const std::string &objectname,
          const std::string &messagename,
          Agent &agent, IncomingMsg &msg, bool verbose)
          : objectname_(objectname), messagename_(messagename),
               verbose_(verbose)
     {
          ++id_;
          
          char buf[10];
          sprintf(buf, "%d", id_);
          thisid_ = buf;

          // register the domain in the local address book
          agent.domainRegister(thisid_, address, port, level);
          
          // send reply to the subscriber with new id
          
          ParamSet params(1);
          params.setString(0, thisid_);
          msg.reply(params);
          
          if (verbose_)
          {
               std::cout << "new subscriber registered:\n"
                    "address: " << address << ", port: " << port
                    << ", level: " << level;
               if (objectname.empty() == false)
                    std::cout << ", object: " << objectname;
               if (messagename.empty() == false)
                    std::cout << ", message: " << messagename;
               std::cout << "\nnew id: " << thisid_ << std::endl;
          }
          
          ok_ = true;
     }

     // returns true if the given id was equal to thisid_
     // otherwise performs no action and returns false
     // after successful unregistration, the relay method
     // should no longer be called
     bool unsubscribe(Agent &agent, const std::string &id)
     {
          if (ok_ == false)
               return false;
          
          if (id == thisid_)
          {
               agent.domainUnregister(thisid_);
               ok_ = false;
               
               if (verbose_)
                    std::cout << "unregistered id: " << thisid_
                         << std::endl;
               
               return true;
          }
          else
               return false;
     }
     
     // if the objectname and messagename match the internal data
     // the params is relayed to the subscriber
     // the subscriber could have subscribed with empty messagename_
     // or empty objectname_ - empty names stored internally
     // match any name that is given as argument
     void relay(Agent &agent,
          const std::string &objectname,
          const std::string &messagename,
          const ParamSet *params)
     {
          if (ok_ == false)
               return;
          
          // do not relay if names do not match
          
          if (objectname_.empty() == false &&
               objectname_ != objectname)
               return;
          
          if (messagename_.empty() == false &&
               messagename_ != messagename)
               return;

          // ignore errors
          try
          {
               if (params != NULL)
               {
                    agent.sendOneWay(thisid_, objectname,
                         messagename, *params);
               }
               else
               {
                    agent.sendOneWay(thisid_, objectname, messagename);
               }

               if (verbose_)
                    std::cout << "relayed to id: " << thisid_
                         << ", object: " << objectname
                         << ", message: " << messagename << std::endl;
          }
          catch (...)
          {
          }
     }
     
private:
     std::string objectname_;
     std::string messagename_;
     std::string thisid_;
     bool ok_;
     bool verbose_;
     
     static int id_;
};

int Subscriber::id_ = 0;


typedef std::list<Subscriber> AllSubscribers;

// the object of this class implements the default object
// which will manage and relay all incoming messages
class EventRelay : public PassiveObject
{
public:
     EventRelay(Agent &agent, AllSubscribers &all, bool verbose)
          :  agent_(agent), all_(all)
     {
     }

     // relay the message to (potentially) every subscriber
     // each sbscriber object will either ignore or resend
     void call(IncomingMsg &msg)
     {
          std::string objectname(msg.getObjectName());
          std::string messagename(msg.getMsgName());
          std::auto_ptr<ParamSet> params(msg.getParameters());

          AllSubscribers::iterator it;
          for (it = all_.begin(); it != all_.end(); ++it)
          {
               it->relay(agent_, objectname, messagename, params.get());
          }
          
          // do not reply nor reject
          
          msg.eat();
     }

private:

     Agent &agent_;
     AllSubscribers &all_;
};


// the object of this class implements the logical YAMI object
// that is the main (managing) event server
// this servant implements only the server's interface
// it's passive so that the agent will automatically notify it
// when any new incoming message arrives
class ESServant : public PassiveObject
{
public:
     ESServant(Agent &agent, AllSubscribers &all, bool verbose)
          : agent_(agent), all_(all), verbose_(verbose)
     {
     }
     
     virtual ~ESServant() {}
     
     void call(IncomingMsg &msg)
     {
          std::string command(msg.getMsgName());
          
          if (verbose_)
               std::cout << "\nevent server got command: "
                    << command << std::endl;
          
          if (command == "subscribe")
               do_subscribe(msg);
          else if (command == "unsubscribe")
               do_unsubscribe(msg);
          else
               if (verbose_)
                    std::cout << "unknown command received" << std::endl;
     }
     
private:
     // copyright
     ESServant(const ESServant &);
     ESServant & operator=(const ESServant &);
     
     // add new listener
     void do_subscribe(IncomingMsg &msg)
     {
          std::auto_ptr<ParamSet> params(msg.getParameters());
          
          if (params.get() == NULL)
          {
               if (verbose_)
                    std::cout << "subscribe without parameters - rejecting"
                         << std::endl;

               msg.reject();
          }
          
          int parcount = params->getParamCount();

          if (parcount < 3 || parcount > 5)
          {
               if (verbose_)
                    std::cout << "subscribe with strange parameters"
                         " - rejecting"     << std::endl;

               msg.reject();
          }
          
          if (params->isASCII(0) == false)
          {
               if (verbose_)
                    std::cout << "subscribe with strange 0 parameter"
                         " - rejecting"     << std::endl;

               msg.reject();
          }
          
          std::string address;
          params->getString(0, address);
          
          if (params->isASCII(1) == false &&
               params->getType(1) != ParamSet::eInt)
          {
               if (verbose_)
                    std::cout << "subscribe with strange 1 parameter"
                         " - rejecting"     << std::endl;

               msg.reject();
          }

          int port;
          
          if (params->getType(1) == ParamSet::eInt)
          {
               port = params->getInt(1);
          }
          else // port is in a string form
          {
               const char *buf = params->getStringBuffer(1);
               if (sscanf(buf, "%d", &port) != 1)
               {
                    if (verbose_)
                         std::cout << "subscribe with strange 1 parameter"
                              " - rejecting"     << std::endl;

                    msg.reject();
               }
          }
          
          if (params->isASCII(2) == false &&
               params->getType(2) != ParamSet::eInt)
          {
               if (verbose_)
                    std::cout << "subscribe with strange 2 parameter"
                         " - rejecting"     << std::endl;

               msg.reject();
          }

          int level;
          
          if (params->getType(2) == ParamSet::eInt)
          {
               level = params->getInt(2);
          }
          else // level is in a string form
          {
               const char *buf = params->getStringBuffer(2);
               if (sscanf(buf, "%d", &level) != 1)
               {
                    if (verbose_)
                         std::cout << "subscribe with strange 2 parameter"
                              " - rejecting"     << std::endl;

                    msg.reject();
               }
          }
          
          std::string objectname;
          std::string messagename;

          if (parcount >= 4)
          {
               if (params->isASCII(3) == false)
               {
                    if (verbose_)
                         std::cout << "subscribe with strange 3 parameter"
                              " - rejecting"     << std::endl;

                    msg.reject();
               }
          
               params->getString(3, objectname);
          
               if (parcount == 5)
               {
                    if (params->isASCII(4) == false)
                    {
                         if (verbose_)
                              std::cout << "subscribe with strange 4 parameter"
                                   " - rejecting"     << std::endl;

                         msg.reject();
                    }
                    
                    params->getString(4, messagename);
               }
          }
          
          all_.push_back(Subscriber(address, port, level,
               objectname, messagename,
               agent_, msg, verbose_));
          
          // note: the new subscriber shakes hand and replies with the
          // new subscription id
     }
     
     // remove the listener
     void do_unsubscribe(IncomingMsg &msg)
     {
          std::auto_ptr<ParamSet> params(msg.getParameters());
          
          if (params.get() == NULL)
          {
               if (verbose_)
                    std::cout << "unsubscribe without parameters - rejecting"
                         << std::endl;

               msg.reject();
          }
          
          int parcount = params->getParamCount();
          
          if (parcount != 1)
          {
               if (verbose_)
                    std::cout << "unsubscribe with strange parameters"
                         " - rejecting"     << std::endl;

               msg.reject();
          }

          if (params->isASCII(0) == false)
          {
               if (verbose_)
                    std::cout << "unsubscribe with strange parameters"
                         " - rejecting"     << std::endl;

               msg.reject();
          }

          std::string id;
          params->getString(0, id);
          
          // find the subscriber with correct id and unsubscribe it
          
          AllSubscribers::iterator it;
          for (it = all_.begin(); it != all_.end(); ++it)
          {
               if (it->unsubscribe(agent_, id))
                    break;
          }
          
          if (it != all_.end())
          {
               // remove the subscriber from the list
               all_.erase(it);
          
               // and shake hand
               msg.reply();
          }
          else // it == all_.end()
          {
               // unknown subscription id
               msg.reject();
          }
     }

     Agent &agent_;
     AllSubscribers &all_;
     bool verbose_;
};


int main(int argc, char **argv)
{
     if (argc != 3 && argc != 4)
     {
          std::cout << "usage:" << std::endl;
          std::cout << "eserver ipport [vq]" << std::endl;
          std::cout << "   ipport is the port on which the server will listen"
               " for messages" << std::endl;
          std::cout << "   v stands for verbose" << std::endl;
          std::cout << "   q stands for quit enable" << std::endl;
          return 0;          
     }

     int ipport = atoi(argv[1]);
     if (ipport <= 0)
     {
          std::cout << "wrong ipport" << std::endl;
          return 0;
     }
     
     bool verbose = false;
     bool quitenabled = false;
     if (argc == 3)
     {
          std::string params(argv[2]);
          if (params.find("v") != std::string::npos)
          {
               verbose = true;
               std::cout << "running in a verbose mode" << std::endl;
          }
          if (params.find("q") != std::string::npos)
          {
               quitenabled = true;
               if (verbose)
                    std::cout << "running with quit enabled" << std::endl;
          }
     }
     
     try
     {
          netInitialize();
          
          {

          Policies policies;
          policies.setConnPoolSize(100);
          policies.setReuseAddr(true);
          policies.setHasWaker(false);
          
          Agent agent(ipport, policies);

          // set up the closing notification object if the
          // 'q' option is enabled
          if (quitenabled)
               agent.objectRegister(es_quit_objectname,
                    Agent::ePolling, NULL);
          
          AllSubscribers subscribers;

          ESServant main_servant(agent, subscribers, verbose);
     
          agent.objectRegister(es_main_objectname,
               Agent::ePassiveSingleThreaded, &main_servant);
          
          if (verbose)
               std::cout << "event service's main servant activated"
                    << std::endl;

          EventRelay relay(agent, subscribers, verbose);
          
          agent.objectRegister(AnyObject,
               Agent::ePassiveSingleThreaded, &relay);
          
          if (verbose)
               std::cout << "relay generic servant activated"
                    << std::endl;
          
          if (verbose)
               std::cout << "waiting for messages..." << std::endl;

          // just go sleep...
          // from now on, all the activity is driven by the agent calling
          // the 'call' method on the servant
          if (quitenabled)
          {
               // wait for any message to the closing object
               // this will wake the server so that it will close itself
               
               std::auto_ptr<IncomingMsg> dummy(
                    agent.getIncoming(es_quit_objectname, true));
               
               if (verbose)
                    std::cout << "got 'quit' command" << std::endl;
          }
          else
          {
               // sleep forever if the 'q' is disabled
               sleep(0);
          }
          
          }
          
          netCleanup();
     }
     catch (const std::exception &e)
     {
          std::cout << e.what() << std::endl;
     }

     return 0;
}
