//
// pserver.cc
//
// this file implements the Property Service, but the
// user is free to use it for any purpose s/he wants
//
// Copyright (C) 2001-2003 Maciej Sobczak,
// student at the Institute of Computer Science,
// Warsaw University of Technology
//
// Permission to copy, use, modify, sell and distribute this software
// is granted provided this copyright notice appears in all copies.
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//

#include "yami++.h"
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <memory>
#include <cctype>

using namespace YAMI;


// this is the name of the object used to close the server
// if the quit option is enabled
const std::string ps_quit_objectname("propertyserver_quit");

// this is the name of the property server's main object
const std::string ps_main_objectname("propertyserver");

// file formatting constants
const int string_width = 70;          // width of string line
const int binary_width = 16;          // width of binary line


// this class encapsulates the value
// of single property
class Value
{
private:

     // generalized value
     class SomeValue
     {
     public:
          virtual void write(std::ostream &os) const = 0;
          virtual void setPSValue(ParamSet *ps, int position) const = 0;
          virtual std::auto_ptr<SomeValue> clone() const = 0;

          virtual ~SomeValue() {}
     };
     
     class StringValue : public SomeValue
     {
     public:
          explicit StringValue(const std::string &value)
               : value_(value)
          {
          }
          
          explicit StringValue(std::istream &is)
          {
               int lines;
               is >> lines;
               std::string dummy;
               std::getline(is, dummy);

               std::string tmp;
               for (int i = 0; i < lines; ++i)
               {
                    std::string tmp2;
                    std::getline(is, tmp2);

                    for (std::string::size_type pos = 0;
                         pos < tmp2.size(); ++pos)
                    {
                         char c = tmp2[pos];
                         if (isalnum(c))
                              tmp += c;
                         else if (c == '.')
                         {
                              char val;
                              val = (tmp2[pos + 1] - '0') * 100 +
                                   (tmp2[pos + 2] - '0') * 10 +
                                   (tmp2[pos + 3] - '0');
                              tmp += val;
                              pos += 3;
                         }
                    }
               }

               value_ = tmp;
          }
          
          void write(std::ostream &os) const
          {
               if (value_.empty())
               {
                    os << 0 << "\n";
                    return;
               }
               
               int lines = 1;
               int charsinline = 0;
               std::string tmp("\t");
               for (std::string::size_type pos = 0;
                    pos < value_.size(); ++pos)
               {
                    char c = value_[pos];
                    if (isalnum(c))
                    {
                         tmp += c;
                         ++charsinline;
                    }
                    else
                    {
                         char buf[5];
                         buf[0] = '.';
                         buf[1] = '0' + c / 100;
                         c %= 100;
                         buf[2] = '0' + c / 10;
                         buf[3] = '0' + c % 10;
                         buf[4] = 0;
                         tmp += buf;
                         charsinline += 4;
                    }
                    if (charsinline >= string_width &&
                         pos < value_.size() - 1)
                    {
                         tmp += "\n\t";
                         ++lines;
                         charsinline = 0;
                    }
               }
               
               os << lines << "\n" << tmp << "\n";
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setString(position, value_);
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new StringValue(*this));
          }
          
     private:
          std::string value_;
     };

#ifndef YAMI_NO_WSTRING     
     class WStringValue : public SomeValue
     {
     public:
          explicit WStringValue(const std::wstring &value)
               : value_(value)
          {
          }
          
          explicit WStringValue(std::istream &is)
          {
               int size;

               is >> size;

               value_.reserve(size);
               
               for (int i = 0; i < size; ++i)
               {
                    int val;
                    is >> val;
                    value_ += (wchar_t)val;
               }

               std::string dummy;
               std::getline(is, dummy);
          }
          
          void write(std::ostream &os) const
          {
               int size = value_.size();
               os << size << "\n";
               
               int lines = (size + binary_width - 1) / binary_width;
               int i = 0;
               for (int l = 0; l < lines; ++l)
               {
                    os << '\t';
                    for (int j = 0; j < binary_width && i < size; ++j, ++i)
                         os << (int)value_[i] << " ";
                    os << "\n";
               }
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setWString(position, value_);
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new WStringValue(*this));
          }
          
     private:
          std::wstring value_;
     };
#endif // YAMI_NO_WSTRING
     
     class IntValue : public SomeValue
     {
     public:
          explicit IntValue(int value)
               : value_(value)
          {
          }
          
          explicit IntValue(std::istream &is)
          {
               is >> value_;

               std::string dummy;
               std::getline(is, dummy);
          }
          
          void write(std::ostream &os) const
          {
               os << value_ << "\n";
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setInt(position, value_);
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new IntValue(*this));
          }
          
     private:
          int value_;
     };

     class DoubleValue : public SomeValue
     {
     public:
          explicit DoubleValue(double value)
               : value_(value)
          {
          }
          
          explicit DoubleValue(std::istream &is)
          {
               is >> value_;

               std::string dummy;
               std::getline(is, dummy);
          }
          
          void write(std::ostream &os) const
          {
               os << value_ << "\n";
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setDouble(position, value_);
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new DoubleValue(*this));
          }
          
     private:
          double value_;
     };
     
     class ByteValue : public SomeValue
     {
     public:
          explicit ByteValue(char value)
               : value_(value)
          {
          }
          
          explicit ByteValue(std::istream &is)
          {
               int ival;

               is >> ival;
               value_ = (char)ival;

               std::string dummy;
               std::getline(is, dummy);
          }
          
          void write(std::ostream &os) const
          {
               os << (int)value_ << "\n";
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setByte(position, value_);
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new ByteValue(*this));
          }
          
     private:
          char value_;
     };
     
     class BinaryValue : public SomeValue
     {
     public:
          BinaryValue(const char *buf, int bufsize)
          {
               value_.assign(buf, buf + bufsize);
          }
          
          explicit BinaryValue(std::istream &is)
          {
               int size;

               is >> size;

               value_.reserve(size);
               
               for (int i = 0; i < size; ++i)
               {
                    int val;
                    is >> val;
                    value_.push_back((char)val);
               }

               std::string dummy;
               std::getline(is, dummy);
          }
          
          void write(std::ostream &os) const
          {
               int size = value_.size();
               os << size << "\n";
               
               int lines = (size + binary_width - 1) / binary_width;
               int i = 0;
               for (int l = 0; l < lines; ++l)
               {
                    os << '\t';
                    for (int j = 0; j < binary_width && i < size; ++j, ++i)
                         os << (int)value_[i] << " ";
                    os << "\n";
               }
          }

          void setPSValue(ParamSet *ps, int position) const
          {
               ps->setBinary(position, &value_[0], value_.size());
          }
          
          std::auto_ptr<SomeValue> clone() const
          {
               return std::auto_ptr<SomeValue>(new BinaryValue(*this));
          }
          
     private:
          std::vector<char> value_;
     };

public:
     Value()
          : type_(ParamSet::eNoType), value_(NULL)
     {
     }
     
     explicit Value(std::istream &is)
     {
          char type;
          is >> type;
          
          switch (type)
          {
          case 's':
               // MSVC++ 6.0 does not implement std::auto_ptr<T>::reset
               // on "normal" compiler, this should read:
               // value_.reset(new StringValue(is));

               {
                    std::auto_ptr<SomeValue> t(new StringValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eString;
               break;
          case 'w':
#ifndef YAMI_NO_WSTRING
               {
                    std::auto_ptr<SomeValue> t(new WStringValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eWString;
#endif // YAMI_NO_WSTRING
               break;
          case 'i':
               {
                    std::auto_ptr<SomeValue> t(new IntValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eInt;
               break;
          case 'd':
               {
                    std::auto_ptr<SomeValue> t(new DoubleValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eDouble;
               break;
          case 'o':
               {
                    std::auto_ptr<SomeValue> t(new ByteValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eByte;
               break;
          case 'b':
               {
                    std::auto_ptr<SomeValue> t(new BinaryValue(is));
                    value_ = t;
               }
               type_ = ParamSet::eBinary;
               break;
          default:
               throw 0;
          }
     }
     
     explicit Value(const std::string &value)
          : type_(ParamSet::eString),
          value_(new StringValue(value))
     {
     }

#ifndef YAMI_NO_WSTRING     
     explicit Value(const std::wstring &value)
          : type_(ParamSet::eWString),
          value_(new WStringValue(value))
     {
     }
#endif // YAMI_NO_WSTRING
     
     explicit Value(int value)
          : type_(ParamSet::eInt),
          value_(new IntValue(value))
     {
     }
     
     explicit Value(double value)
          : type_(ParamSet::eDouble),
          value_(new DoubleValue(value))
     {
     }
     
     explicit Value(char value)
          : type_(ParamSet::eByte),
          value_(new ByteValue(value))
     {
     }
     
     explicit Value(const char *buf, int bufsize)
          : type_(ParamSet::eBinary),
          value_(new BinaryValue(buf, bufsize))
     {
     }

     Value(const Value &v)
          : type_(v.type_),
          value_(v.value_.get() ?
               v.value_->clone() :
               std::auto_ptr<SomeValue>(NULL))
     {
     }
     
     Value& operator=(const Value &v)
     {
          Value tmp(v);
          swap(tmp);
          return *this;
     }
     
     void write(std::ostream &os) const
     {
          switch (type_)
          {
          case ParamSet::eString:
               os << "\ts ";
               break;
          case ParamSet::eWString:
#ifndef YAMI_NO_WSTRING
               os << "\tw ";
#endif // YAMI_NO_WSTRING
               break;
          case ParamSet::eInt:
               os << "\ti ";
               break;
          case ParamSet::eDouble:
               os << "\td ";
               break;
          case ParamSet::eByte:
               os << "\to ";
               break;
          case ParamSet::eBinary:
               os << "\tb ";
               break;
          default:
               throw 0;
          }

          value_->write(os);
     }
     
     // instead of normal operator=
     void swap(Value &v)
     {
          std::auto_ptr<SomeValue> tmp1 = value_;
          value_ = v.value_;
          v.value_ = tmp1;
          
          enum ParamSet::eType tmp2 = type_;
          type_ = v.type_;
          v.type_ = tmp2;
     }

     void setPSValue(ParamSet *ps, int position) const
     {
          value_->setPSValue(ps, position);
     }
     
private:
     enum ParamSet::eType type_;
     std::auto_ptr<SomeValue> value_;
};

// this class encapsulates the collection
// of properties for one object
class Properties
{
private:
     // this is the type of collection
     typedef std::map<std::string, Value> properties_type;
     
public:
     typedef properties_type::iterator iterator;
     typedef properties_type::const_iterator const_iterator;

     Properties() {}
          
     explicit Properties(std::istream &is)
     {
          int propcount;
          is >> propcount;
          
          std::string dummy;
          std::getline(is, dummy);

          for (int i = 0; i < propcount; ++i)
          {
               std::string propname;
               std::getline(is, propname);

               Value v(is);
               properties_.insert(
                    properties_type::value_type(propname, v));
          }

          std::getline(is, dummy);
     }

     iterator begin()
     {
          return properties_.begin();
     }
     
     iterator end()
     {
          return properties_.end();
     }
     
     int size() const
     {
          return properties_.size();
     }
     
     iterator find(const std::string &propname)
     {
          return properties_.find(propname);
     }

     template <class T>
     iterator set_property(const std::string &propname,
          const T &value)
     {
          std::pair<iterator, bool> p = properties_.insert(
               properties_type::value_type(propname, Value(value)));
          if (p.second == false)
               p.first->second = Value(value);
          return p.first;
     }
     
     iterator set_property(const std::string &propname,
          const char *buf, int bufsize)
     {
          std::pair<iterator, bool> p = properties_.insert(
               properties_type::value_type(propname, Value(buf, bufsize)));
          if (p.second == false)
               p.first->second = Value(buf, bufsize);
          return p.first;
     }
          
     void remove(const std::string &propname)
     {
          properties_.erase(propname);
     }

     void write(std::ostream &os) const
     {
          os << properties_.size() << "\n";
          
          const_iterator it(properties_.begin());
          const_iterator end(properties_.end());
          
          for (; it != end; ++it)
          {
               os << it->first << "\n";
               it->second.write(os);
          }
          
          os << "---------------------------------\n";
     }

private:
     properties_type properties_;
};

// this class encapsulates
// the collection of objects
class Objects
{
private:
     // this is the type of collection
     typedef std::map<std::string, Properties> objects_type;

public:
     typedef objects_type::iterator iterator;
     typedef objects_type::const_iterator const_iterator;
     
     explicit Objects(std::istream &is)
     {
          while (1)
          {
               std::string objname;
               std::getline(is, objname);
               if (objname.empty())
                    break;

               Properties props(is);
               objects_.insert(
                    objects_type::value_type(objname, props));
          }
     }

     iterator begin()
     {
          return objects_.begin();
     }
     
     iterator end()
     {
          return objects_.end();
     }
     
     int size() const
     {
          return objects_.size();
     }
     
     iterator find(const std::string &objname)
     {
          return objects_.find(objname);
     }

     iterator new_object(const std::string &objname)
     {
          std::pair<iterator, bool> p = objects_.insert(
               objects_type::value_type(objname, Properties()));
          if (p.second == false)
               p.first->second = Properties();
          return p.first;
     }
          
     void remove(const std::string &objname)
     {
          objects_.erase(objname);
     }

     void write(std::ostream &os) const
     {
          const_iterator it(objects_.begin());
          const_iterator end(objects_.end());
          
          for (; it != end; ++it)
          {
               os << it->first << "\n";
               it->second.write(os);
          }
     }
     
private:
     // copyright
     Objects(const Objects &);
     Objects& operator=(const Objects &);
     
     objects_type objects_;
};


// the object of this class implements the logical YAMI object
// that is the actual server of the property service's data
// this servant implements only the server's interface
// it's passive so that the agent will automatically notify it
// when any new incoming message arrives
class PSServant : public PassiveObject
{
public:
     PSServant(Objects &objects, const std::string &filename,
          bool verbose, bool readonly)
          : filename_(filename), objects_(objects),
          verbose_(verbose), readonly_(readonly)
     {
     }
     
     virtual ~PSServant() {}
     
     // crucial method in passive servants
     // this will be called by the agent
     void call(IncomingMsg &msg)
     {
          std::string command(msg.getMsgName());
          
          if (verbose_)
               std::cout << std::endl
                    << "property_server got command: "
                    << command << std::endl;
          
          if (command == "new")
               do_new(msg);
          else if (command == "remove")
               do_remove(msg);
          else if (command == "list")
               do_list(msg);
          else if (command == "write")
               do_write(msg);
          else
               if (verbose_)
                    std::cout << "unknown command received" << std::endl;
     }
     
private:
     // copyright
     PSServant(const PSServant &);
     PSServant& operator=(const PSServant &);
     
     // add new entry
     void do_new(IncomingMsg &msg)
     {
          if (readonly_)
          {
               if (verbose_)
                    std::cout << "'new' command not allowed with readonly"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          std::auto_ptr<YAMI::ParamSet> parset(msg.getParameters());
          if (parset.get() == NULL)
          {
               if (verbose_)
                    std::cout << "'new' command with no parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          int valuecount = parset->getParamCount();
          if (valuecount != 1)
          {
               if (verbose_)
                    std::cout << "'add' command with strange parameters"
                         " - rejecting"     << std::endl;
               msg.reject();
               return;
          }

          // the only parameter should be string
          // (here we enforce that it is
          // also plain string, not Unicode)
          if (parset->isASCII(0) != true)
          {
               if (verbose_)
                    std::cout << "'add' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          // create new entry
          std::string objname;
          parset->getString(0, objname);
          objects_.new_object(objname);

          if (verbose_)
               std::cout << "adding new object: " << objname << std::endl;

          // shake hand
          msg.reply();
     }
     
     // remove an entry
     void do_remove(IncomingMsg &msg)
     {
          if (readonly_)
          {
               if (verbose_)
                    std::cout << "'remove' command not allowed with readonly"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          // we expect that the 'remove' message brings only one
          // parameter of type string with the name of the object
          
          std::auto_ptr<YAMI::ParamSet> parset(msg.getParameters());
          if (parset.get() == NULL)
          {
               if (verbose_)
                    std::cout << "'remove' command with no parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          int valuecount = parset->getParamCount();
          if (valuecount != 1)
          {
               if (verbose_)
                    std::cout << "'remove' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          if (parset->isASCII(0) != true)
          {
               if (verbose_)
                    std::cout << "'remove' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          // remove the object
          std::string objname;
          parset->getString(0, objname);

          if (verbose_)
               std::cout << "removing object: " << objname << std::endl;

          objects_.remove(objname);

          // shake hand
          msg.reply();
     }

     // list all the objects
     void do_list(IncomingMsg &msg)
     {
          // we do not care what parameters it brings
          
          if (verbose_)
               std::cout << "listing objects" << std::endl;

          int objectcount = objects_.size();
          
          if (objectcount == 0)
          {
               if (verbose_)
                    std::cout << "there are no objects" << std::endl;

               // empty response means "no objects"
               msg.reply();
          }
          else
          {
               ParamSet retparset(objectcount);
               
               Objects::iterator it(objects_.begin());
               Objects::iterator itend(objects_.end());
          
               for (int i = 0; it != itend; ++it, ++i)
               {
                    retparset.setString(i, it->first);
               }
               
               // respond with all the names
               msg.reply(retparset);
          }
     }
          
     // write the data to the file
     void do_write(IncomingMsg &msg)
     {
          if (readonly_)
          {
               if (verbose_)
                    std::cout << "'write' command not allowed with readonly"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          std::ofstream ofile(filename_.c_str());
          objects_.write(ofile);

          // shake hand
          msg.reply();
     }
     
     std::string filename_;
     Objects &objects_;
     bool verbose_;
     bool readonly_;
};


// the object of this class implements all the logical YAMI objects
// that have some properties defined
// it's passive so that the agent will automatically notify it
// when any new incoming message arrives
class ObjectServant : public PassiveObject
{
public:
     ObjectServant(Objects &objects, bool verbose, bool readonly)
          : objects_(objects), verbose_(verbose), readonly_(readonly)
     {
     }
     
     // crucial method in passive servants
     // this will be called by the agent
     void call(IncomingMsg &msg)
     {
          std::string objname(msg.getObjectName());
          
          // find if this object is in the collection
          Objects::iterator it(objects_.find(objname));
          if (it == objects_.end())
          {
               // no such object
               if (verbose_)
                    std::cout << "unknown object: " << objname << std::endl;
               
               msg.reject();
               return;
          }

          if (verbose_)
               std::cout << "object " << objname << ": ";
                    
          std::string command(msg.getMsgName());
          if (command == "set")
               do_set(it, msg);
          else if (command == "get")
               do_get(it, msg);
          else if (command == "remove")
               do_remove(it, msg);
          else if (command == "list")
               do_list(it, msg);
          else
               if (verbose_)
                    std::cout << "unknown command : "
                         << command << std::endl;
     }
     
private:
     // copyright
     ObjectServant(const ObjectServant &);
     ObjectServant& operator=(ObjectServant &);

     void do_set(Objects::iterator it, IncomingMsg &msg)
     {
          if (readonly_)
          {
               if (verbose_)
                    std::cout << "'set' not allowed with readonly"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          std::auto_ptr<ParamSet> paramset(msg.getParameters());
          
          if (paramset.get() == NULL)
          {
               if (verbose_)
                    std::cout << "'set' command with no parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          // check if the parameter set contains exactly two parameters
          int paramcount = paramset->getParamCount();
          if (paramcount != 2)
          {
               if (verbose_)
                    std::cout << "'set' with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          if (paramset->isASCII(0) != true)
          {
               if (verbose_)
                    std::cout << "'set' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          std::string propname;
          paramset->getString(0, propname);
          
          if (verbose_)
               std::cout << "'set' property: " << propname << std::endl;

          enum ParamSet::eType type = paramset->getType(1);
          switch (type)
          {
          case ParamSet::eString:
               {
                    std::string value;
                    paramset->getString(1, value);
                    it->second.set_property(propname, value);
               }
               break;
          case ParamSet::eWString:
#ifndef YAMI_NO_WSTRING
               {
                    std::wstring value;
                    paramset->getWString(1, value);
                    it->second.set_property(propname, value);
               }
#endif // YAMI_NO_WSTRING
               break;
          case ParamSet::eInt:
               {
                    int value(paramset->getInt(1));
                    it->second.set_property(propname, value);
               }
               break;
          case ParamSet::eDouble:
               {
                    double value(paramset->getDouble(1));
                    it->second.set_property(propname, value);
               }
               break;
          case ParamSet::eByte:
               {
                    char value;
                    it->second.set_property(propname, value);
               }
               break;
          case ParamSet::eBinary:
               {
                    int size(paramset->getBinarySize(1));
                    std::vector<char> value(size);
                    paramset->getBinaryValue(1, &value[0]);
                    it->second.set_property(propname, &value[0], size);
               }
               break;
          default:
               throw 0;
          }
          
          // shake hand
          msg.reply();
     }
     
     void do_get(Objects::iterator it, IncomingMsg &msg)
     {
          // we expect that the 'get' message brings non-empty
          // list of parameters with the names of the properties
          
          // the response will contain the properties' values,
          // in the order corresponding to the property names
          // in the query, but only when all the names can be find
          
          // if the query is for one property only, the empty response
          // means that the property is not defined
          // if the query is for more than one property and at least
          // one of the names cannot be found, the message will be rejected
          
          std::auto_ptr<YAMI::ParamSet> paramset(msg.getParameters());
          if (paramset.get() == NULL)
          {
               if (verbose_)
                    std::cout << "'get' command with no parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          int paramcount = paramset->getParamCount();

          // check the names in query
          
          int i;
          for (i = 0; i < paramcount; ++i)
          {
               if (paramset->isASCII(i) != true)
               {
                    if (verbose_)
                         std::cout << "'get' command with strange parameters"
                              " - rejecting" << std::endl;
                    msg.reject();
                    return;
               }

               // look up this property
               std::string propname;
               paramset->getString(i, propname);
               Properties::iterator pit(it->second.find(propname));
               
               if (pit == it->second.end())
               {
                    if (verbose_)
                         std::cout << "no property " << propname
                              << std::endl;

                    if (paramcount == 1)
                    {
                         // empty response means "no such property"
                         msg.reply();
                         return;
                    }
                    else
                    {
                         // reject if the query was for more than one name
                         msg.reject();
                         return;
                    }
               }
          }

          // here, we are sure that all the names in the query are defined

          std::auto_ptr<ParamSet> retparset(new ParamSet(paramcount));

          for (i = 0; i < paramcount; ++i)
          {
               std::string propname;
               paramset->getString(i, propname);
               Properties::iterator pit(it->second.find(propname));
               
               if (verbose_)
                    std::cout << "'get' property: " << propname << std::endl;

               pit->second.setPSValue(retparset.get(), i);
          }

          // respond with all the values
          msg.reply(*retparset);
     }

     void do_remove(Objects::iterator it, IncomingMsg &msg)
     {
          if (readonly_)
          {
               if (verbose_)
                    std::cout << "'remove' allowed with readonly"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }
          
          // we expect that the 'remove' message brings only one
          // parameter of type string with the name of the property
          
          std::auto_ptr<YAMI::ParamSet> parset(msg.getParameters());
          if (parset.get() == NULL)
          {
               if (verbose_)
                    std::cout << "'remove' command with no parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          int valuecount = parset->getParamCount();
          if (valuecount != 1)
          {
               if (verbose_)
                    std::cout << "'remove' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          if (parset->isASCII(0) != true)
          {
               if (verbose_)
                    std::cout << "'remove' command with strange parameters"
                         " - rejecting" << std::endl;
               msg.reject();
               return;
          }

          // remove the property
          std::string propname;
          parset->getString(0, propname);

          if (verbose_)
               std::cout << "removing property: " << propname << std::endl;

          it->second.remove(propname);

          // shake hand
          msg.reply();
     }
          
     void do_list(Objects::iterator it, IncomingMsg &msg)
     {
          // we do not care what parameters it brings
          
          if (verbose_)
               std::cout << "listing properties" << std::endl;

          int propertycount = it->second.size();
          
          if (propertycount == 0)
          {
               if (verbose_)
                    std::cout << "there are no properties" << std::endl;

               // empty response means "no properties"
               msg.reply();
          }
          else
          {
               ParamSet retparset(propertycount);
               
               Properties::iterator pit(it->second.begin());
               Properties::iterator pitend(it->second.end());
          
               for (int i = 0; pit != pitend; ++pit, ++i)
               {
                    retparset.setString(i, pit->first);
               }
               
               // respond with all the names
               msg.reply(retparset);
          }
     }
     
     Objects &objects_;
     bool verbose_;
     bool readonly_;
};
          

int main(int argc, char **argv)
{
     if (argc != 3 && argc != 4)
     {
          std::cout << "usage:" << std::endl;
          std::cout << "pserver filename ipport [vrq]" << std::endl;
          std::cout << "   filename is a file that stores the server's data"
               << std::endl;
          std::cout << "   ipport is the port on which the server will listen"
               " for messages" << std::endl;
          std::cout << "   v stands for verbose" << std::endl;
          std::cout << "   r stands for readonly" << std::endl;
          std::cout << "   q stands for quit enable" << std::endl;
          return 0;          
     }

     int ipport = atoi(argv[2]);
     if (ipport <= 0)
     {
          std::cout << "wrong ipport" << std::endl;
          return 0;
     }
     
     bool verbose = false;
     bool readonly = false;
     bool quitenabled = false;
     if (argc == 4)
     {
          std::string params(argv[3]);
          if (params.find("v") != std::string::npos)
          {
               verbose = true;
               std::cout << "running in a verbose mode" << std::endl;
          }
          if (params.find("r") != std::string::npos)
          {
               readonly = true;
               if (verbose)
                    std::cout << "running in a readonly mode" << std::endl;
          }
          if (params.find("q") != std::string::npos)
          {
               quitenabled = true;
               if (verbose)
                    std::cout << "running with quit enabled" << std::endl;
          }
     }
     
     std::string filename(argv[1]);
          
     std::ifstream ifile(filename.c_str());
     if (!ifile)
     {
          std::cout << "cannot open file " << filename << std::endl;
          return -1;
     }
     
     Objects objects(ifile);
     ifile.close();
     
     try
     {
          netInitialize();
          
          {

          Policies policies;
          policies.setConnPoolSize(20);
          policies.setReuseAddr(true);
          policies.setHasWaker(false);
          
          Agent agent(ipport, policies);

          // set up the closing notification object if the
          // 'q' option is enabled
          if (quitenabled)
               agent.objectRegister(ps_quit_objectname,
                    Agent::ePolling, NULL);
          
          PSServant main_servant(objects, filename, verbose, readonly);
     
          agent.objectRegister(ps_main_objectname,
               Agent::ePassiveSingleThreaded, &main_servant);
          
          if (verbose)
               std::cout << "property service's main servant activated"
                    << std::endl;
          
          ObjectServant objects_servant(objects, verbose, readonly);
          
          agent.objectRegister(AnyObject,
               Agent::ePassiveSingleThreaded, &objects_servant);
          
          if (verbose)
               std::cout << "objects generic servant activated"
                    << std::endl;
          
          if (verbose)
               std::cout << "waiting for messages..." << std::endl;

          // just go sleep...
          // from now on, all the activity is driven by the agent calling
          // the 'call' method on the servant
          if (quitenabled)
          {
               // wait for any message to the closing object
               // this will wake the server so that it will close itself
               
               std::auto_ptr<IncomingMsg> dummy(
                    agent.getIncoming(ps_quit_objectname, true));
               
               if (verbose)
                    std::cout << "got 'quit' command" << std::endl;
          }
          else
          {
               // sleep forever if the 'q' is disabled
               sleep(0);
          }
          
          }
          
          netCleanup();
     }
     catch (const std::exception &e)
     {
          std::cout << e.what() << std::endl;
     }

     return 0;
}
