//
// cppgen.cc
//
// definition of the C++ generator
//
// Copyright (C) 2001-2003 Maciej Sobczak,
// student at the Institute of Computer Science,
// Warsaw University of Technology
//
// Permission to copy, use, modify, sell and distribute this software
// is granted provided this copyright notice appears in all copies.
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//

#include "cppgen.h"
#include <fstream>
#include <string>

using namespace std;

void CPPGenerator::generate(const YDL &ydl, const string &corename,
          const options &opts, ostream &err)
{
     string hsuffix;
     options::const_iterator it = opts.find("hsuffix");
     if (it == opts.end())
     {
          hsuffix = ".h";
     }
     else
     {
          hsuffix = "." + it->second;
     }
     
     string cppsuffix;
     it = opts.find("cppsuffix");
     if (it == opts.end())
     {
          cppsuffix = ".cc";
     }
     else
     {
          cppsuffix = "." + it->second;
     }
     
     string namespc;
     it = opts.find("namespace");
     if (it == opts.end())
     {
          namespc = "";
     }
     else
     {
          namespc = it->second;
     }

     eBinStrForm stringinform;
     it = opts.find("stringin");
     if (it == opts.end())
     {
          stringinform = eStdString;
     }
     else
     {
          if (it->second == "ptr")
               stringinform = ePointer;
          else if (it->second == "string")
               stringinform = eStdString;
          else
          {
               err << "ydl: String input form not recognized: "
                    << it->second << "\n";
               throw 0;
          }
     }

     eBinStrForm stringoutform;
     it = opts.find("stringout");
     if (it == opts.end())
     {
          stringoutform = eStdString;
     }
     else
     {
          if (it->second == "malloc")
               stringoutform = eMalloc;
          else if (it->second == "new")
               stringoutform = eNew;
          else if (it->second == "string")
               stringoutform = eStdString;
          else
          {
               err << "ydl: String output form not recognized: "
                    << it->second << "\n";
               throw 0;
          }
     }

     eBinStrForm bininform;
     it = opts.find("binin");
     if (it == opts.end())
     {
          bininform = eStdVector;
     }
     else
     {
          if (it->second == "ptr")
               bininform = ePointer;
          else if (it->second == "vector")
               bininform = eStdVector;
          else
          {
               err << "ydl: Binary input form not recognized: "
                    << it->second << "\n";
               throw 0;
          }
     }

     eBinStrForm binoutform;
     it = opts.find("binout");
     if (it == opts.end())
     {
          binoutform = eStdVector;
     }
     else
     {
          if (it->second == "malloc")
               binoutform = eMalloc;
          else if (it->second == "new")
               binoutform = eNew;
          else if (it->second == "vector")
               binoutform = eStdVector;
          else
          {
               err << "ydl: Binary output form not recognized: "
                    << it->second << "\n";
               throw 0;
          }
     }

     {
          string filename(corename + "_client" + hsuffix);
          ofstream file(filename.c_str());
          if (file)
          {
               generateClientHeader(ydl, file, namespc, filename,
                    stringinform, stringoutform,
                    bininform, binoutform, err);
          }
          else
          {
               err << "ydl: " << filename
                    << ": Cannot open file for writing\n";
               throw 0;
          }
     }

     {
          string headername(corename + "_client" + hsuffix);
          string filename(corename + "_client" + cppsuffix);
          ofstream file(filename.c_str());
          if (file)
          {
               generateClientDefinitions(ydl, file, namespc, headername,
                    stringinform, stringoutform,
                    bininform, binoutform, err);
          }
          else
          {
               err << "ydl: " << filename
                    << ": Cannot open file for writing\n";
               throw 0;
          }
     }

     {
          string filename(corename + "_server" + hsuffix);
          ofstream file(filename.c_str());
          if (file)
          {
               generateServerHeader(ydl, file, namespc, filename,
                    stringinform, stringoutform, bininform, binoutform, err);
          }
          else
          {
               err << "ydl: " << filename
                    << ": Cannot open file for writing\n";
               throw 0;
          }
     }

     {
          string headername(corename + "_server" + hsuffix);
          string filename(corename + "_server" + cppsuffix);
          ofstream file(filename.c_str());
          if (file)
          {
               generateServerDefinitions(ydl, file, namespc, headername,
                    stringinform, stringoutform,
                    bininform, binoutform, err);
          }
          else
          {
               err << "ydl: " << filename
                    << ": Cannot open file for writing\n";
               throw 0;
          }
     }
}

string CPPGenerator::getUpper(const string &name)
{
     string upper(name);
     for (string::iterator it = upper.begin(); it != upper.end(); ++it)
     {
          if (isalpha(*it))
          {
               *it = toupper(*it);
          }
          else if (!isalnum(*it))
          {
               *it = '_';
          }
     }
     
     return upper;
}

void CPPGenerator::doIndent(ostream &os, int indent)
{
     for (int i = 0; i != indent; ++i)
     {
          os << "    ";
     }
}

void CPPGenerator::generateParams(const Message &msg, ostream &os,
     eBinStrForm stringinform, eBinStrForm stringoutform,
     eBinStrForm bininform, eBinStrForm binoutform,
     int indent, ostream &err)
{
     os << "(";

     bool paramspresent = (msg.input.empty() == false ||
          msg.output.empty() == false);     

     if (paramspresent)
     {
          os << "\n";
     }

     // first, all input parameters
     
     for (vector<Param>::const_iterator inpar = msg.input.begin();
          inpar != msg.input.end(); ++inpar)
     {
          switch (inpar->type)
          {
          case eString:
               switch (stringinform)
               {
               case ePointer:
                    doIndent(os, indent);
                    os << "    const char *" << inpar->name;
                    break;
               case eStdString:
                    doIndent(os, indent);
                    os << "    const std::string &"
                         << inpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringinform)
               {
               case ePointer:
                    doIndent(os, indent);
                    os << "    const wchar_t *" << inpar->name;
                    break;
               case eStdString:
                    doIndent(os, indent);
                    os << "    const std::wstring &"
                         << inpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               doIndent(os, indent);
               os << "    int " << inpar->name;
               break;
          case eDouble:
               doIndent(os, indent);
               os << "    double " << inpar->name;
               break;
          case eByte:
               doIndent(os, indent);
               os << "    char " << inpar->name;
               break;
          case eBinary:
               switch (bininform)
               {
               case ePointer:
                    doIndent(os, indent);
                    os << "    const char *" << inpar->name
                         << ", " << "size_t " << inpar->name
                         << "Size";
                    break;
               case eStdVector:
                    doIndent(os, indent);
                    os << "    const std::vector<char> &"
                         << inpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          }
          
          if (inpar + 1 != msg.input.end())
               os << ",\n";
     }
     
     // then, all the output parameters
     
     if (msg.input.empty() == false && msg.output.empty() == false)
          os << ",\n";
     
     for (vector<Param>::const_iterator outpar = msg.output.begin();
          outpar != msg.output.end(); ++outpar)
     {
          switch (outpar->type)
          {
          case eString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    doIndent(os, indent);
                    os << "    char *&" << outpar->name;
                    break;
               case eStdString:
                    doIndent(os, indent);
                    os << "    std::string &"
                         << outpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    doIndent(os, indent);
                    os << "    wchar_t *&" << outpar->name;
               break;
               case eStdString:
                    doIndent(os, indent);
                    os << "    std::wstring &"
                         << outpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               doIndent(os, indent);
               os << "    int &" << outpar->name;
               break;
          case eDouble:
               doIndent(os, indent);
               os << "    double &" << outpar->name;
               break;
          case eByte:
               doIndent(os, indent);
               os << "    char &" << outpar->name;
               break;
          case eBinary:
               switch (binoutform)
               {
               case eMalloc:
               case eNew:
                    doIndent(os, indent);
                    os << "    char *&" << outpar->name
                         << ", " << "size_t &" << outpar->name
                         << "Size";
                    break;
               case eStdVector:
                    doIndent(os, indent);
                    os << "    std::vector<char> &"
                         << outpar->name;
                    break;
               default:
                    throw 0;
               }
               break;
          }
     
          if (outpar + 1 != msg.output.end())
               os << ",\n";
     }

     if (paramspresent)
     {
          os << "\n";
          doIndent(os, indent);
     }
     
     os << ")";
}

void CPPGenerator::generateClientHeader(const YDL &ydl, ostream &os,
     const string &namespc, const string &filename,
     eBinStrForm stringinform, eBinStrForm stringoutform,
     eBinStrForm bininform, eBinStrForm binoutform,
     ostream &err)
{
     os << "//\n";
     os << "// this file has been generated by the YDL compiler\n";
     os << "//\n";
     os << "// it is supposed to be #included by the client code\n";
     os << "// that wants to send messages according to YDL definitions\n";
     os << "//\n";
     os << "\n";
     
     string upper = getUpper(filename) + "_INCLUDED";
     os << "#ifndef " << upper << "\n";
     os << "#define " << upper << "\n";

     os << "\n";
     os << "#include <string>\n";
     os << "#include <vector>\n";
     os << "\n";
     
     // generate definition for the exceptions
     
     os << "// common exceptions\n";
     os << "\n";
     
     os << "namespace YAMI\n";
     os << "{\n";
     os << "\n";
     os << "class BadResponse : public std::runtime_error\n";
     os << "{\n";
     os << "public:\n";
     os << "    BadResponse()\n"
          "        : std::runtime_error(\"incorrect response\")\n"
          "    {\n"
          "    }\n";
     os << "};\n";
     os << "\n";
     os << "class TimeOut : public std::runtime_error\n";
     os << "{\n";
     os << "public:\n";
     os << "    TimeOut()\n"
          "        : std::runtime_error(\"the message was timed out\")\n"
          "    {\n"
          "    }\n";
     os << "};\n";
     os << "\n";
     os << "class Reject : public std::runtime_error\n";
     os << "{\n";
     os << "public:\n";
     os << "    Reject()\n"
          "        : std::runtime_error(\"the message was rejected\")\n"
          "    {\n"
          "    }\n";
     os << "};\n";
     os << "\n";
     os << "} // namespace YAMI\n";
     os << "\n";

     if (namespc.empty() == false)
     {
          os << "namespace " << namespc << "\n";
          os << "{\n";
          os << "\n";
     }
     
     // generate declarations for every interface
     
     os << "// YDL interfaces\n";
     os << "\n";
     
     for (YDL::const_iterator interf = ydl.begin();
          interf != ydl.end(); ++interf)
     {
          string classname(interf->name);
          
          os << "class " << classname << "\n";
          os << "{\n";
          os << "public:\n";
          os << "\n";
          os << "    // constructors\n";
          os << "\n";
          os << "    " << classname << "(::YAMI::Agent &agent, "
               "const std::string &domainname,\n"
               "        const std::string &objectname);\n";
          os << "    " << classname << "(::YAMI::Agent &agent, "
               "const char *domainname,\n"
               "        const char *objectname);\n";
          os << "\n";
          os << "    // rebind functionality\n";
          os << "\n";
          os << "    void rebind(::YAMI::Agent &agent, "
               "const std::string &domainname,\n"
               "        const std::string &objectname);\n";
          os << "    void rebind(::YAMI::Agent &agent, "
               "const char *domainname,\n"
               "        const char *objectname);\n";
          os << "\n";
          os << "    // timeout functionality\n";
          os << "\n";
          os << "    void setTimeOut(int timeout);\n";
          os << "\n";
          os << "    // messages\n";
          os << "\n";
          
          for (vector<Message>::const_iterator msg = interf->messages.begin();
               msg != interf->messages.end(); ++msg)
          {
               os << "    void " << msg->name;

               generateParams(*msg, os, stringinform, stringoutform,
                    bininform, binoutform, 1, err);
               
               os << ";\n";
               os << "\n";
          }
          
          os << "private:\n";
          os << "    // copyright\n";
          os << "    " << classname << "(const " << classname << " &);\n";
          os << "    " << classname << " & operator=(const " << classname
               << " &);\n";
          os << "\n";
          os << "    ::YAMI::Agent *agent_;\n";
          os << "    const char *domainname_;\n";
          os << "    const char *objectname_;\n";
          os << "    int timeout_;\n";
          os << "};\n";
          os << "\n";
     }
     
     if (namespc.empty() == false)
     {
          os << "} // namespace " << namespc << "\n";
          os << "\n";
     }
     
     os << "#endif // " << upper << "\n";
}

// this function generates code that pushes parameters
// to the ParamSet object
// generated code is used on the client side
void CPPGenerator::generateParamSetFeedCli(ostream &os,
     const vector<Param> &params, const char *paramsname,
     eBinStrForm stringinform, eBinStrForm bininform,
     std::ostream &err)
{
     int i = 0;
     for (vector<Param>::const_iterator itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          os << "    " << paramsname << ".";
          
          switch (itpar->type)
          {
          case eString:
               switch (stringinform)
               {
               case ePointer:
                    os << "setString(" << i << ", " << itpar->name << ");\n";
                    break;
               case eStdString:
                    os << "setString(" << i << ", " << itpar->name
                         << ".c_str());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringinform)
               {
               case ePointer:
                    os << "setWString(" << i << ", " << itpar->name << ");\n";
                    break;
               case eStdString:
                    os << "setWString(" << i << ", " << itpar->name
                         << ".c_str());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               os << "setInt(" << i << ", " << itpar->name << ");\n";
               break;
          case eDouble:
               os << "setDouble(" << i << ", " << itpar->name << ");\n";
               break;
          case eByte:
               os << "setByte(" << i << ", " << itpar->name << ");\n";
               break;
          case eBinary:
               switch (bininform)
               {
               case ePointer:
                    os << "setBinary(" << i << ", " << itpar->name
                         << ", " << itpar->name << "Size);\n";
                    break;
               case eStdVector:
                    os << "setBinary(" << i << ", &" << itpar->name << "[0]"
                         << ", " << itpar->name << ".size());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          }
     }
}

void CPPGenerator::generateSizeDecl(ostream &os, bool &alreadyGenerated)
{
     if (alreadyGenerated == false)
     {
          os << "size_t ";
          alreadyGenerated = true;
     }
}

// this function generates code that reads ParamSet object
// and write the data to appropriate function parameters
// generated code is used at the client side
void CPPGenerator::generateParamSetReadCli(ostream &os,
     const vector<Param> &params, const char *paramsname,
     eBinStrForm stringoutform, eBinStrForm binoutform,
     std::ostream &err)
{
     os << "    // check the number of parameters\n"
          "    if (" << paramsname << ".get() == NULL)\n"
          "    {\n"
          "        throw ::YAMI::BadResponse();\n"
          "    }\n";
     os << "    if (" << paramsname << "->getParamCount() != "
          << params.size() << ")\n"
          "    {\n"
          "        throw ::YAMI::BadResponse();\n"
          "    }\n";
     os << "\n";
     os << "    // check the types of parameters\n";
     os << "\n";

     int i = 0;
     vector<Param>::const_iterator itpar;
     for (itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          os << "    if (" << paramsname << "->getType(" << i
               << ") != ::YAMI::ParamSet::";

          switch (itpar->type)
          {
          case eString:
               os << "eString";
               break;
          case eWString:
               os << "eWString";
               break;
          case eInt:
               os << "eInt";
               break;
          case eDouble:
               os << "eDouble";
               break;
          case eByte:
               os << "eByte";
               break;
          case eBinary:
               os << "eBinary";
               break;
          }

          os << ")\n"
               "    {\n"
               "        throw ::YAMI::BadResponse();\n"
               "    }\n";
     }
     
     os << "\n";
     os << "    // OK, read the parameters\n";
     os << "\n";
     
     bool sizeAlreadyGenerated = false;
     for (i = 0, itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          switch (itpar->type)
          {
          case eString:
               switch (stringoutform)
               {
               case eMalloc:
                    os << "    ";
                    generateSizeDecl(os, sizeAlreadyGenerated);
                    os << "size = outparams->getStringLength("
                         << i << ");\n";
                    os << "    " << itpar->name
                         << " = (char*)malloc(size + 1);\n";
                    os << "    malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getStringValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eNew:
                    os << "    ";
                    generateSizeDecl(os, sizeAlreadyGenerated);
                    os << "size = outparams->getStringLength("
                         << i << ");\n";
                    os << "    " << itpar->name << " = new char[size + 1];\n";
                    os << "    new_pointer_handle<char> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getStringValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eStdString:
                    os << "    outparams->getString(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringoutform)
               {
               case eMalloc:
                    os << "    ";
                    generateSizeDecl(os, sizeAlreadyGenerated);
                    os << "size = outparams->getWStringLength("
                         << i << ");\n";
                    os << "    " << itpar->name
                         << " = (wchar_t*)malloc((size + 1) * "
                         "sizeof(wchar_t));\n";
                    os << "    malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getWStringValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eNew:
                    os << "    ";
                    generateSizeDecl(os, sizeAlreadyGenerated);
                    os << "size = outparams->getWStringLength("
                         << i << ");\n";
                    os << "    " << itpar->name
                         << " = new wchar_t[size + 1];\n";
                    os << "    new_pointer_handle<wchar_t> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getWStringValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eStdString:
                    os << "    outparams->getWString(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               os << "    " << itpar->name << " = outparams->getInt("
                    << i << ");\n";
               os << "\n";
               break;
          case eDouble:
               os << "    " << itpar->name << " = outparams->getDouble("
                    << i << ");\n";
               os << "\n";
               break;
          case eByte:
               os << "    " << itpar->name << " = outparams->getByte("
                    << i << ");\n";
               os << "\n";
               break;
          case eBinary:
               switch (binoutform)
               {
               case eMalloc:
                    os << "    " << itpar->name
                         << "Size = outparams->getBinarySize("
                         << i << ");\n";
                    os << "    " << itpar->name
                         << " = (char*)malloc(" << itpar->name << "Size);\n";
                    os << "    malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getBinaryValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eNew:
                    os << "    " << itpar->name
                         << "Size = outparams->getBinarySize("
                         << i << ");\n";
                    os << "    " << itpar->name << " = new char["
                         << itpar->name << "Size];\n";
                    os << "    new_pointer_handle<char> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    os << "    outparams->getBinaryValue(" << i << ", "
                         << itpar->name << ");\n";
                    os << "\n";
                    break;
               case eStdVector:
                    os << "    ";
                    generateSizeDecl(os, sizeAlreadyGenerated);
                    os << "size = outparams->getBinarySize(" << i << ");\n";
                    os << "    " << itpar->name << ".resize(size);\n";
                    os << "    outparams->getBinaryValue(" << i << ", &"
                         << itpar->name << "[0]);\n";
                    os << "\n";
                    break;
               default:
                    throw 0;
               }
               break;
          }
     }

     // and at the end, release the handlers
     
     for (itpar = params.begin();
          itpar != params.end(); ++itpar)
     {
          switch (itpar->type)
          {
          case eString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    os << "    " << itpar->name << "_handle.release();\n";
                    break;
               default:
                    break;
               }
               break;
          case eWString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    os << "    " << itpar->name << "_handle.release();\n";
                    break;
               default:
                    break;
               }
               break;
          case eBinary:
               switch (binoutform)
               {
               case eMalloc:
               case eNew:
                    os << "    " << itpar->name << "_handle.release();\n";
                    break;
               default:
                    break;
               }
               break;
          default:
               break;
          }
     }
}

void CPPGenerator::generateHandlers(ostream &os, ostream &err)
{
     os << "struct malloc_pointer_handle\n"
          "{\n"
          "    malloc_pointer_handle(void *p) : p_(p) {}\n"
          "    ~malloc_pointer_handle() { free(p_); }\n"
          "    void release() { p_ = NULL; }\n"
          "    void *p_;\n"
          "};\n"
          "\n"
          "template <typename T>\n"
          "struct new_pointer_handle\n"
          "{\n"
          "    new_pointer_handle(T *p) : p_(p) {}\n"
          "    ~new_pointer_handle() { delete [] p_; }\n"
          "    void release() { p_ = NULL; }\n"
          "    T *p_;\n"
          "};\n"
          "\n";
}

void CPPGenerator::generateClientDefinitions(const YDL &ydl, ostream &os,
     const string &namespc, const string &headername,
     eBinStrForm stringinform, eBinStrForm stringoutform,
     eBinStrForm bininform, eBinStrForm binoutform,
     ostream &err)
{
     os << "//\n";
     os << "// this file has been generated by the YDL compiler\n";
     os << "//\n";
     os << "// it is supposed to be compiled and linked "
          "with the client code\n";
     os << "// that wants to send messages according to YDL definitions\n";
     os << "//\n";
     os << "\n";
     os << "#include \"yami++.h\"\n";
     os << "#include \"yamierrors.h\"\n";
     os << "#include \"" << headername << "\"\n";
     os << "\n";
     
     os << "// helper definitions\n"
          "\n"
          "namespace\n"
          "{\n"
          "\n";

     // generate helper function for checking message status
     
     os << "// check the message status - common function\n"
          "// this function throws if status != eReplied\n"
          "void checkStatus(::YAMI::Message::eStatus status)\n"
          "{\n"
          "    if (status == ::YAMI::Message::eTimedOut)\n"
          "    {\n"
          "        throw ::YAMI::TimeOut();\n"
          "    }\n"
          "    else if (status == ::YAMI::Message::eRejected ||\n"
          "        status == ::YAMI::Message::eUnknownObj ||\n"
          "        status == ::YAMI::Message::eRejectByAgent)\n"
          "    {\n"
          "        throw ::YAMI::Reject();\n"
          "    }\n"
          "    else if (status == ::YAMI::Message::eOverflow)\n"
          "    {\n"
          "        throw ::YAMI::Overflow();\n"
          "    }\n"
          "    else if (status == ::YAMI::Message::eNetError)\n"
          "    {\n"
          "        throw ::YAMI::OSError(YAMINETERROR);\n"
          "    }\n"
          "}\n"
          "\n";
     
     generateHandlers(os, err);
     
     os << "} // namespace\n"
          "\n";
     
     string spcqualif;
     if (namespc.empty() == true)
     {
          spcqualif = "";
     }
     else
     {
          spcqualif = namespc + "::";
     }

     // generate definitions for every interface

     for (YDL::const_iterator interf = ydl.begin();
          interf != ydl.end(); ++interf)
     {
          string classname(interf->name);
          
          os << "// definitions for the " << classname << " interface\n";
          os << "\n";
          os << "// constructors\n";
          os << "\n";
          os << spcqualif << classname << "::" << classname
               << "(::YAMI::Agent &agent,\n"
               "        const std::string &domainname,\n"
               "        const std::string &objectname)\n";
          os << "    : agent_(&agent), domainname_(domainname.c_str()),\n"
               "        objectname_(objectname.c_str()), timeout_(0)\n"
               "{\n"
               "}\n";
          os << "\n";
          os << spcqualif << classname << "::" << classname
               << "(::YAMI::Agent &agent,\n"
               "        const char *domainname,\n"
               "        const char *objectname)\n";
          os << "    : agent_(&agent), domainname_(domainname),\n"
               "        objectname_(objectname), timeout_(0)\n"
               "{\n"
               "}\n";
          os << "\n";
          os << "// rebind functionality\n";
          os << "\n";
          os << "void " << spcqualif << classname << "::"
               << "rebind(::YAMI::Agent &agent,\n"
               "        const std::string &domainname,\n"
               "        const std::string &objectname)\n"
               "{\n"
               "    agent_ = &agent;\n"
               "    domainname_ = domainname.c_str();\n"
               "    objectname_ = objectname.c_str();\n"
               "}\n";
          os << "\n";
          os << "void " << spcqualif << classname << "::"
               << "rebind(::YAMI::Agent &agent,\n"
               "        const char *domainname,\n"
               "        const char *objectname)\n"
               "{\n"
               "    agent_ = &agent;\n"
               "    domainname_ = domainname;\n"
               "    objectname_ = objectname;\n"
               "}\n";
          os << "\n";
          os << "// timeout functionality\n";
          os << "\n";
          os << "void " << spcqualif << classname << "::"
               << "setTimeOut(int timeout)\n"
               "{\n"
               "    timeout_ = timeout;\n"
               "}\n";
          os << "\n";
          os << "// messages\n";
          os << "\n";
          
          for (vector<Message>::const_iterator msg = interf->messages.begin();
               msg != interf->messages.end(); ++msg)
          {
               os << "void " << spcqualif << classname << "::" << msg->name;

               generateParams(*msg, os, stringinform, stringoutform,
                    bininform, binoutform, 0, err);
               
               os << "\n";
               os << "{\n";
               
               if (msg->input.empty() == true)
               {
                    os << "    // no input parameters\n";
                    os << "\n";
                    
                    if (msg->oneway == true)
                    {
                         os << "    // send one-way message\n";
                         os << "\n";
                         os << "    agent_->sendOneWay(\n"
                              "        domainname_, objectname_, \""
                              << msg->name << "\");\n";
                    }
                    else
                    {
                         os << "    // send the message\n";
                         os << "\n";
                         os << "    std::auto_ptr< ::YAMI::Message> message(\n"
                              "        agent_->send("
                              "domainname_, objectname_, \"" 
                              << msg->name << "\"));\n";
                    }
               }
               else
               {
                    os << "    // prepare the input parameters\n";
                    os << "\n";
                    
                    os << "    ::YAMI::ParamSet inparams("
                         << msg->input.size() << ");\n";
                    
                    generateParamSetFeedCli(os, msg->input, "inparams",
                         stringinform, bininform, err);
                    
                    os << "\n";
                    if (msg->oneway == true)
                    {
                         os << "    // send one-way message\n";
                         os << "\n";
                         os << "    agent_->sendOneWay(\n"
                              "        domainname_, objectname_, \""
                              << msg->name << "\" , inparams);\n";
                    }
                    else
                    {
                         os << "    // send the message\n";
                         os << "\n";
                         os << "    std::auto_ptr< ::YAMI::Message> message(\n"
                              "        agent_->send("
                              "domainname_, objectname_, \"" << msg->name
                              << "\", inparams));\n";
                    }
               }

               if (msg->oneway == false)
               {
                    os << "\n";
                    os << "    // set the timeout, if required\n";
                    os << "\n";
                    os << "    if (timeout_ > 0)\n"
                         "    {\n"
                         "        message->setTimeOut(timeout_);\n"
                         "    }\n";
                    os << "\n";
                    os << "    // wait for the response\n";
                    os << "\n";
                    os << "    message->wait();\n";
                    os << "\n";
                    os << "    ::YAMI::Message::eStatus status = "
                         "message->getStatus();\n";
                    os << "    checkStatus(status); "
                         "// throws if status != eReplied\n";
                    os << "\n";
                    os << "    // status == eReplied\n";
                    os << "\n";
                    os << "    // get the reply\n";
                    os << "\n";
                    
                    // put the "RAII" wrappers
                    
                    // TODO: finish here
                    // there should be releasers inistead of try/catch blocks
                    
                    os << "    std::auto_ptr< ::YAMI::ParamSet> outparams("
                         "message->getResponse());\n";
                    os << "\n";
                    
                    if (msg->output.empty() == true)
                    {
                         // no out params, check for it
                         
                         os << "    // there should be no output params\n";
                         os << "    if (outparams.get() != NULL)\n"
                              "    {\n"
                              "        throw ::YAMI::BadResponse();\n"
                              "    }\n";
                    }
                    else
                    {
                         // process the output params
                         
                         generateParamSetReadCli(os, msg->output, "outparams",
                              stringoutform, binoutform, err);
                    }
               }

               os << "}\n";
               os << "\n";
          }
     }
}

void CPPGenerator::generateServerHeader(const YDL &ydl, ostream &os,
     const string &namespc, const string &filename,
     eBinStrForm stringinform, eBinStrForm stringoutform,
     eBinStrForm bininform, eBinStrForm binoutform,
     ostream &err)
{
     os << "//\n";
     os << "// this file has been generated by the YDL compiler\n";
     os << "//\n";
     os << "// it is supposed to be #included by the server code\n";
     os << "// that wants to receive messages according to YDL definitions\n";
     os << "//\n";
     os << "\n";
     
     string upper = getUpper(filename) + "_INCLUDED";
     os << "#ifndef " << upper << "\n";
     os << "#define " << upper << "\n";

     os << "\n";
     os << "#include <string>\n";
     os << "#include <vector>\n";
     os << "\n";

     if (namespc.empty() == false)
     {
          os << "namespace " << namespc << "\n";
          os << "{\n";
          os << "\n";
     }
     
     // generate declarations for every skeleton
     
     os << "// YDL skeletons\n";
     os << "\n";
     
     for (YDL::const_iterator interf = ydl.begin();
          interf != ydl.end(); ++interf)
     {
          string classname(interf->name + "_Skel");
          
          os << "class " << classname << " : public ::YAMI::PassiveObject\n";
          os << "{\n";
          os << "private:\n";
          os << "\n";
          os << "    // messages\n";
          os << "\n";
          
          for (vector<Message>::const_iterator msg = interf->messages.begin();
               msg != interf->messages.end(); ++msg)
          {
               os << "    virtual void " << msg->name;
               
               generateParams(*msg, os, stringinform, stringoutform,
                    bininform, binoutform, 1, err);
               
               os << " = 0;\n";
               os << "\n";
          }
          
          os << "    // this method by default does nothing,\n"
               "    // which results in message rejection\n"
               "    // override it if you want different behavior\n";
          os << "    virtual void unknownMessage"
               << "(::YAMI::IncomingMsg &incomingmsg) {}\n";
          os << "\n";
          os << "    // this is used for actual dispatching\n";
          os << "    void call(::YAMI::IncomingMsg &incomingmsg);\n";
          os <<"};\n";
          os << "\n";
     }
     
     if (namespc.empty() == false)
     {
          os << "} // namespace " << namespc << "\n";
          os << "\n";
     }
     
     os << "#endif // " << upper << "\n";
}

// this function generates code that reads ParamSet object
// and write the data to appropriate function parameters
// generated code is used at the server side
void CPPGenerator::generateParamSetReadSvr(ostream &os,
     const vector<Param> &params, const char *paramsname,
     eBinStrForm stringinform, eBinStrForm bininform,
     std::ostream &err)
{
     os << "        // check the number of parameters\n"
          "        if (" << paramsname << ".get() == NULL)\n"
          "        {\n"
          "            incomingmsg.reject();\n"
          "            return;\n"
          "        }\n";
     os << "        if (" << paramsname << "->getParamCount() != "
          << params.size() << ")\n"
          "        {\n"
          "            incomingmsg.reject();\n"
          "            return;\n"
          "        }\n";
     os << "\n";
     os << "        // check the types of parameters\n";
     os << "\n";

     int i = 0;
     vector<Param>::const_iterator itpar;
     for (itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          os << "        if (" << paramsname << "->getType(" << i
               << ") != ::YAMI::ParamSet::";

          switch (itpar->type)
          {
          case eString:
               os << "eString";
               break;
          case eWString:
               os << "eWString";
               break;
          case eInt:
               os << "eInt";
               break;
          case eDouble:
               os << "eDouble";
               break;
          case eByte:
               os << "eByte";
               break;
          case eBinary:
               os << "eBinary";
               break;
          }

          os << ")\n"
               "        {\n"
               "            incomingmsg.reject();\n"
               "            return;\n"
               "        }\n";
     }
     
     os << "\n";
     os << "        // OK, read the parameters\n";
     os << "\n";
     
     for (i = 0, itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          switch (itpar->type)
          {
          case eString:
               switch (stringinform)
               {
               case ePointer:
                    os << "        const char *" << itpar->name
                         << " = " << paramsname << "->getStringBuffer("
                         << i << ");\n";
                    break;
               case eStdString:
                    os << "        std::string " << itpar->name << ";\n"
                         "        " << paramsname << "->getString("
                         << i << ", " << itpar->name << ");\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringinform)
               {
               case ePointer:
                    os << "        const wchar_t *" << itpar->name
                         << " = " << paramsname << "->getWStringBuffer("
                         << i << ");\n";
                    break;
               case eStdString:
                    os << "        std::wstring " << itpar->name << ";\n"
                         "        " << paramsname << "->getWString("
                         << i << ", " << itpar->name << ");\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               os << "        int " << itpar->name << " = " << paramsname
                    << "->getInt(" << i << ");\n";
               break;
          case eDouble:
               os << "        double " << itpar->name << " = " << paramsname
                    << "->getDouble(" << i << ");\n";
               break;
          case eByte:
               os << "        char " << itpar->name << " = " << paramsname
                    << "->getByte(" << i << ");\n";
               break;
          case eBinary:
               switch (bininform)
               {
               case ePointer:
                    os << "        const char *" << itpar->name
                         << " = " << paramsname << "->getBinaryBuffer("
                         << i << ");\n";
                    os << "        size_t " << itpar->name << "Size = "
                         << paramsname << "->getBinarySize(" << i << ");\n";
                    break;
               case eStdVector:
                    os << "        std::vector<char> " << itpar->name << "("
                         << paramsname << "->getBinarySize(" << i << "));\n"
                         "        " << paramsname << "->getBinaryValue("
                         << i << ", &" << itpar->name << "[0]);\n";
                    break;
               default:
                    throw 0;
               }
               break;
          }
     }
}

// this function generates code for RAII (well, not exactly)
// wrappers for memory allocated by the called servant
void CPPGenerator::generateHandlersSvr(ostream &os,
     const vector<Param> &params,
     eBinStrForm stringoutform, eBinStrForm binoutform,
     ostream &err)
{
     bool atleastone = false;
     for (vector<Param>::const_iterator itpar = params.begin();
          itpar != params.end(); ++itpar)
     {
          switch (itpar->type)
          {
          case eString:
               switch (stringoutform)
               {
               case eMalloc:
                    os << "        malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eNew:
                    os << "        new_pointer_handle<char> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eStdString:
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringoutform)
               {
               case eMalloc:
                    os << "        malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eNew:
                    os << "        new_pointer_handle<wchar_t> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eStdString:
                    break;
               default:
                    throw 0;
               }
               break;
          case eBinary:
               switch (binoutform)
               {
               case eMalloc:
                    os << "        malloc_pointer_handle " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eNew:
                    os << "        new_pointer_handle<char> " << itpar->name
                         << "_handle(" << itpar->name << ");\n";
                    atleastone = true;
                    break;
               case eStdVector:
                    break;
               default:
                    throw 0;
               }
               break;
          default:
               // other types do not require wrappers
               break;
          }
     }
     
     if (atleastone == true)
          os << "\n";
}

// this function generates code that pushes parameters
// to the ParamSet object
// generated code is used on the server side
void CPPGenerator::generateParamSetFeedSvr(ostream &os,
     const vector<Param> &params, const char *paramsname,
     eBinStrForm stringoutform, eBinStrForm binoutform,
     ostream &err)
{
     int i = 0;
     for (vector<Param>::const_iterator itpar = params.begin();
          itpar != params.end(); ++itpar, ++i)
     {
          os << "        " << paramsname << ".";
          
          switch (itpar->type)
          {
          case eString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    os << "setString(" << i << ", " << itpar->name << ");\n";
                    break;
               case eStdString:
                    os << "setString(" << i << ", " << itpar->name
                         << ".c_str());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eWString:
               switch (stringoutform)
               {
               case eMalloc:
               case eNew:
                    os << "setWString(" << i << ", " << itpar->name << ");\n";
                    break;
               case eStdString:
                    os << "setWString(" << i << ", " << itpar->name
                         << ".c_str());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          case eInt:
               os << "setInt(" << i << ", " << itpar->name << ");\n";
               break;
          case eDouble:
               os << "setDouble(" << i << ", " << itpar->name << ");\n";
               break;
          case eByte:
               os << "setByte(" << i << ", " << itpar->name << ");\n";
               break;
          case eBinary:
               switch (binoutform)
               {
               case eMalloc:
               case eNew:
                    os << "setBinary(" << i << ", " << itpar->name
                         << ", " << itpar->name << "Size);\n";
                    break;
               case eStdVector:
                    os << "setBinary(" << i << ", &" << itpar->name << "[0]"
                         << ", " << itpar->name << ".size());\n";
                    break;
               default:
                    throw 0;
               }
               break;
          }
     }
}

void CPPGenerator::generateServerDefinitions(const YDL &ydl, ostream &os,
     const string &namespc, const string &headername,
     eBinStrForm stringinform, eBinStrForm stringoutform,
     eBinStrForm bininform, eBinStrForm binoutform,
     ostream &err)
{
     os << "//\n";
     os << "// this file has been generated by the YDL compiler\n";
     os << "//\n";
     os << "// it is supposed to be compiled and linked "
          "with the server code\n";
     os << "// that wants to receive messages according to YDL definitions\n";
     os << "//\n";
     os << "\n";
     os << "#include \"yami++.h\"\n";
//     os << "#include \"yamierrors.h\"\n";
     os << "#include \"" << headername << "\"\n";
     os << "#include <string>\n";
     os << "\n";

     os << "// helper definitions\n"
          "\n"
          "namespace\n"
          "{\n"
          "\n";

     generateHandlers(os, err);
     
     os << "} // namespace\n"
          "\n";
     
     string spcqualif;
     if (namespc.empty() == true)
     {
          spcqualif = "";
     }
     else
     {
          spcqualif = namespc + "::";
     }

     // generate definitions for every interface

     for (YDL::const_iterator interf = ydl.begin();
          interf != ydl.end(); ++interf)
     {
          string classname(interf->name + "_Skel");
          
          os << "// dispatcher for the " << classname << " interface\n";
          os << "\n";
          os << "void " << spcqualif << classname
               << "::call(::YAMI::IncomingMsg &incomingmsg)\n"
               "{\n"
               "    // dispatch on the message name\n"
               "\n"
               "    std::string msgname(incomingmsg.getMsgName());\n";
          
          bool firstif = true;     
          for (vector<Message>::const_iterator msg = interf->messages.begin();
               msg != interf->messages.end(); ++msg)
          {
               os << "    ";
               
               if (firstif == false)
               {
                    os << "else ";
               }
               else
               {
                    firstif = false;
               }
               
               os << "if (msgname == \"" << msg->name << "\")\n"
                    "    {\n";
               os << "        std::auto_ptr< ::YAMI::ParamSet> inparams("
                    "incomingmsg.getParameters());\n"
                    "\n";
               
               if (msg->input.empty() == true)
               {
                    // no in params, check for it
                    
                    os << "        // there should be no input params\n"
                         "        if (inparams.get() != NULL)\n"
                         "        {\n"
                         "            incomingmsg.reject();\n"
                         "            return;\n"
                         "        }\n"
                         "\n";
               }
               else
               {
                    // process the input params
                    
                    os << "        // process the input params\n";
                    
                    generateParamSetReadSvr(os, msg->input, "inparams",
                         stringinform, bininform, err);
                    
                    os << "\n";
               }
                    
               // declare the output params, if any
               
               if (msg->oneway == true || msg->output.empty() == true)
               {
                    os << "        // no output parameters\n"
                         "\n";
               }
               else
               {
                    os << "        // declare out parameters\n"
                         "\n";
                    
                    for (vector<Param>::const_iterator
                         itpar = msg->output.begin();
                         itpar != msg->output.end();
                         ++itpar)
                    {
                         switch (itpar->type)
                         {
                         case eString:
                              switch (stringoutform)
                              {
                              case eMalloc:
                              case eNew:
                                   os << "        char *"
                                        << itpar->name << ";\n";
                                   break;
                              case eStdString:
                                   os << "        std::string "
                                        << itpar->name << ";\n";
                                   break;
                              default:
                                   throw 0;
                              }
                              break;
                         case eWString:
                              switch (stringoutform)
                              {
                              case eMalloc:
                              case eNew:
                                   os << "        wchar_t *"
                                        << itpar->name << ";\n";
                                   break;
                              case eStdString:
                                   os << "        std::wstring "
                                        << itpar->name << ";\n";
                                   break;
                              default:
                                   throw 0;
                              }
                              break;
                         case eInt:
                              os << "        int "
                                   << itpar->name << ";\n";
                              break;
                         case eDouble:
                              os << "        double "
                                   << itpar->name << ";\n";
                              break;
                         case eByte:
                              os << "        char "
                                   << itpar->name << ";\n";
                              break;
                         case eBinary:
                              switch (binoutform)
                              {
                              case eMalloc:
                              case eNew:
                                   os << "        char *"
                                        << itpar->name << ";\n"
                                        "        size_t "
                                        << itpar->name << "Size;\n";
                                   break;
                              case eStdVector:
                                   os << "        std::vector<char> "
                                        << itpar->name << ";\n";
                                   break;
                              default:
                                   throw 0;
                              }
                         }
                    }
                    os << "\n";
               }
               
               os << "        // perform the physical call\n"
                    "\n";
               os << "        " << msg->name << "(";

               bool parameters = (msg->input.empty() == false ||
                                    msg->output.empty() == false);
               if (parameters)
                    os << "\n";

               for (vector<Param>::const_iterator inpar = msg->input.begin();
                    inpar != msg->input.end(); ++inpar)
               {
                    switch (inpar->type)
                    {
                    case eString:
                    case eWString:
                    case eInt:
                    case eDouble:
                    case eByte:
                         os << "            " << inpar->name;
                         break;
                    case eBinary:
                         switch (bininform)
                         {
                         case ePointer:
                              os << "            " << inpar->name << ", "
                                   << inpar->name << "Size";
                              break;
                         case eStdVector:
                              os << "            " << inpar->name;
                              break;
                         default:
                              throw 0;
                         }
                         break;
                    }
                    
                    if (inpar + 1 != msg->input.end())
                         os << ",\n";
               }
               
               if (msg->input.empty() == false && msg->output.empty() == false)
                    os << ",\n";
               
               for (vector<Param>::const_iterator outpar = msg->output.begin();
                    outpar != msg->output.end(); ++outpar)
               {
                    switch (outpar->type)
                    {
                    case eString:
                    case eWString:
                    case eInt:
                    case eDouble:
                    case eByte:
                         os << "            " << outpar->name;
                         break;
                    case eBinary:
                         switch (binoutform)
                         {
                         case eMalloc:
                         case eNew:
                              os << "            " << outpar->name << ", "
                                   << outpar->name << "Size";
                              break;
                         case eStdVector:
                              os << "            " << outpar->name;
                              break;
                         default:
                              throw 0;
                         }
                         break;
                    }
                    
                    if (outpar + 1 != msg->output.end())
                         os << ",\n";
               }

               if (parameters)
                    os << "\n"
                         "        );\n";
               else
                    os << ");\n";
               
               os << "\n";
               
               // put the "RAII" wrappers
               
               generateHandlersSvr(os, msg->output, stringoutform,
                    binoutform, err);
               
               // process the output
               
               if (msg->oneway == true)
               {
                    os << "        // one-way message, no reply\n"
                         "        incomingmsg.eat();\n";
               }
               else
               {
                    if (msg->output.empty() == true)
                    {
                         os << "        // there is no output param\n"
                              "\n"
                              "        incomingmsg.reply();\n";
                    }
                    else
                    {
                         os << "        // pack the output params "
                              "and send them\n"
                              "\n"
                              "        ::YAMI::ParamSet outparams("
                              << msg->output.size() << ");\n";
                         
                         generateParamSetFeedSvr(os, msg->output, "outparams",
                              stringoutform, binoutform, err);
                         
                         os << "\n"
                              "        incomingmsg.reply(outparams);\n";
                    }
               }
     
               os << "    }\n";
          }
          os << "    else\n"
               "    {\n"
               "        // call the unknown handler\n"
               "        // (by default, reject the message)\n"
               "\n"
               "        unknownMessage(incomingmsg);\n"
               "    }\n";
          
          os << "}\n"
               "\n";
     }
}
