//
// parser.cc
//
// definition of the parser
//
// Copyright (C) 2001-2003 Maciej Sobczak,
// student at the Institute of Computer Science,
// Warsaw University of Technology
//
// Permission to copy, use, modify, sell and distribute this software
// is granted provided this copyright notice appears in all copies.
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//

#include "parser.h"
#include <cstring>
#include <cctype>

using namespace std;

const char * Parser::opentoken[]    = {"(", "{", "[", "begin"};
const char * Parser::closetoken[]   = {")", "}", "]", "end"};
const char * Parser::intoken[]      = {"<", "<<", "in"};
const char * Parser::outtoken[]     = {">", ">>", "out"};
const char * Parser::onewaytoken[]  = {"oneway"};
const char * Parser::endtoken[]     = {"."};
const char * Parser::commenttoken[] = {"//", ";", "#"};

// note:
// the order here conforms to the order in eType in ydl.h file
const char * Parser::typetoken[]    =
          {"string", "wstring", "int", "double", "byte", "binary"};

Parser::Parser(istream &input, ostream &errors)
     : linenumber_(0), in_(input), err_(errors)
{
     // start reading the file
     p_ = currentline_.begin();
     bool ok = getNextChar(true);
     if (ok == false)
     {
          printMsg("Cannot read from input file");
          throw 0;
     }
     
     while (1 == 1)
     {
          ok = getInterface();
          if (ok == false)
               throw 0;
          
          ok = getEnd();
          if (ok == true)
               break;
     }
}

bool Parser::getParams(vector<Param> &params)
{
     bool ok = getOpen();
     if (ok == false)
     {
          printMsg("Open bracket expected");
          showError();
          return false;
     }
     
     // read the parameters
     
     while (1 == 1)
     {
          ok = getType();
          if (ok == false)
          {
               printMsg("Type specifier expected");
               showError();
               return false;
          }
     
          int types = sizeof(typetoken)/sizeof(const char*);
          int i;
          eType t;
          for (i = 0; i != types; ++i)
          {
               if (currenttoken_ == typetoken[i])
               {
                    t = static_cast<eType>(i);
                    break;
               }
          }
          if (i == types)
          {
               // no type matches
     
               printMsg("Type specifier expected");
               showError();
               return false;
          }
          
          // after the type specifier, there should be a parameter name
          
          ok = getName();
          if (ok == false)
          {
               printMsg("Parameter name expected");
               showError();
               return false;
          }
          
          Param newparameter(t, currenttoken_);
          
          params.push_back(newparameter);
          
          // if there is a comma, read more parameters
          // if no, this is end of the list
          
          ok = skipWhites();
          if (ok == false)
               return false;
     
          if (*p_ == ',')
          {
               ok = getNextChar(true);
               if (ok == false)
                    return false;
               continue;
          }
          else
          {
               break;
          }
     }
     
     ok = getClose();
     if (ok == false)
     {
          printMsg("Close bracket expected");
          showError();
          return false;
     }
     
     return true;
}

bool Parser::getInOutOneWayEnd(Message &msg)
{
     bool ok = getIn();
     if (ok == false)
     {
          // no 'in', read the rest, if present
          
          return getOutOneWayEnd(msg);
     }
     else
     {
          // there was 'in' keyword, read the params
               
          ok = getParams(msg.input);
          if (ok == false)
               return false;
          
          // after the input params, there can be out, oneway or end
          
          return getOutOneWayEnd(msg);
     }
}

bool Parser::getOutOneWayEnd(Message &msg)
{
     bool ok = getOut();
     if (ok == false)
     {
          // no 'out' read the rest, if present
          
          return getOneWayEnd(msg);
     }
     else
     {
          // there was 'out' keyword, read the params
          
          ok = getParams(msg.output);
          if (ok == false)
               return false;
          
          // after the output params, read the end
          // (the message cannot be oneway if it has output params)
          
          ok = getEnd();
          if (ok == false)
          {
               printMsg("End expected");
               showError();
               return false;
          }
          
          return true;
     }
}

bool Parser::getOneWayEnd(Message &msg)
{
     bool ok = getOneWay();
     if (ok == false)
     {
          ok = getEnd();
          if (ok == false)
          {
               printMsg("Token not recognized");
               showError();
               return false;
          }
     }
     else
     {
          msg.oneway = true;
          
          ok = getEnd();
          if (ok == false)
          {
               printMsg("End expected");
               showError();
               return false;
          }
     }
     
     return true;
}

bool Parser::getInterface()
{
     bool ok = getName();
     if (ok == false)
     {
          printMsg("Interface name expected");
          showError();
          return false;
     }

     Interface newinterface(currenttoken_);

     ok = getOpen();
     if (ok == false)
     {
          printMsg("Open bracket expected");
          showError();
          return false;
     }

     // read messages
     // note: the list of messages can be empty
     
     while (1 == 1)
     {
          ok = getClose();
          if (ok == true)
               break;
     
          // ok, there should be a message here
          
          ok = getName();
          if (ok == false)
          {
               printMsg("Message name expected");
               showError();
               return false;
          }
          
          Message newmessage(currenttoken_);
          
          // after the name, there can be any of:
          // - in keyword
          // - out keyword
          // - oneway indicator
          // - end
          // and they can be omitted in the same order, except the end

          ok = getInOutOneWayEnd(newmessage);
          if (ok == false)
               return false;
          
          newinterface.messages.push_back(newmessage);
     }

     ydl_.push_back(newinterface);
          
     return true;
}

void Parser::printMsg(const char *msg)
{
     err_ << "ydl: " << msg << " at line " << linenumber_ << '\n';
}

void Parser::showError()
{
     err_ << currentline_ << '\n';
     for (string::iterator i = currentline_.begin(); i != p_; ++i)
     {
          if (*i == '\t')
               err_ << '\t';
          else
               err_ << ' ';
     }
     err_ << "^\n";
}

bool Parser::readLine()
{
     if (getline(in_, currentline_))
     {
          ++linenumber_;
          p_ = currentline_.begin();
          return true;
     }
     else
          return false;
}

bool Parser::getNextChar(bool skipnewlines)
{
     if (p_ + 1 == currentline_.end())
          ++p_;

     if (p_ == currentline_.end())
     {
          // end of the line
          
          if (skipnewlines == true)
          {
               while (p_ == currentline_.end())
               {
                    if (readLine() == false)
                         return false;
               }
          }
          else
          {
               return false;
          }
     }
     else
     {
          // in the middle of the line
          ++p_;
     }

     return true;
}

bool Parser::beginsWith(string::iterator i1, string::iterator i2,
     const char *p)
{
     while (*p && i1 != i2)
     {
          if (*p != *i1)
               return false;
          ++p;
          ++i1;
     }
     if (i1 == i2 && *p)
          return false;

     return true;
}

bool Parser::skipWhites()
{
     while (1 == 1)
     {
          while (p_ == currentline_.end() || isspace(*p_))
          {
               if (getNextChar(true) == false)
                    return false;
          }

          // now, it can be a comment
          string::iterator end = currentline_.end();
          int comments = sizeof(commenttoken)/sizeof(const char*);
          int i;
          for (i = 0; i != comments; ++i)
          {
               if (beginsWith(p_, end, commenttoken[i]))
               {
                    p_ = currentline_.end();
                    break; // for
               }
          }
          
          if (i == comments)
          {
               // p_ was not at the beginning of a comment
               break; // while
          }
     }

     return true;
}

bool Parser::getName()
{
     if (skipWhites() == false)
          return false;

     if (!isalpha(*p_) && *p_ != '_')
     {
          return false;
     }
     
     currenttoken_ = "";
     currenttoken_ += *p_;
     if (getNextChar(false) == false)
          return false;
     while (isalnum(*p_) || *p_ == '_')
     {
          currenttoken_ += *p_;
          if (getNextChar(false) == false)
          {
               // it does not have to be error
               break;
          }
     }
     
     return true;
}

bool Parser::getOpen()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(opentoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, opentoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(opentoken[i]);
               return true;
          }
     }
     
     // opentoken not found
     return false;
}

bool Parser::getClose()
{
     if (skipWhites() == false)
          return false;

     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(closetoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, closetoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(closetoken[i]);
               return true;
          }
     }
     
     // closetoken not found
     return false;
}

bool Parser::getIn()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(intoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, intoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(intoken[i]);
               return true;
          }
     }
     
     // intoken not found
     return false;
}

bool Parser::getOut()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(outtoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, outtoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(outtoken[i]);
               return true;
          }
     }
     
     // outtoken not found
     return false;
}

bool Parser::getType()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(typetoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, typetoken[i]))
          {
               // currenttoken_ is needed here
               int len = strlen(typetoken[i]);
               currenttoken_.assign(p_, p_ + len);
               p_ += len;
               return true;
          }
     }
     
     // typetoken not found
     return false;
}

bool Parser::getOneWay()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(onewaytoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, onewaytoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(onewaytoken[i]);
               return true;
          }
     }
     
     // onewaytoken not found
     return false;
}

bool Parser::getEnd()
{
     if (skipWhites() == false)
          return false;
     
     string::iterator end = currentline_.end();
     for (int i = 0; i != sizeof(endtoken)/sizeof(const char*); ++i)
     {
          if (beginsWith(p_, end, endtoken[i]))
          {
               // currenttoken_ is not needed here
               p_ += strlen(endtoken[i]);
               return true;
          }
     }
     
     // endtoken not found
     return false;
}
